#
# This file is part of gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function, division

import random
import threading
import unittest

import gruvi
from gruvi.hub import get_hub
from support import UnitTest


class LockTests(object):
    # Shared tests for Lock and RLock

    Lock = None
    nfibers = None
    nthreads = None

    def test_basic(self):
        # Lock and unlock the rlock once.
        lock = gruvi.RLock()
        lock.acquire()
        self.assertTrue(lock.locked())
        lock.release()
        self.assertFalse(lock.locked())

    def test_context_manager(self):
        # Ensure that a lock can be used as a context manager.
        lock = self.Lock()
        with lock:
            self.assertTrue(lock.locked())
        self.assertFalse(lock.locked())

    def test_lock_order(self):
        # Locks are fair and are granted in order.
        lock = self.Lock()
        order = []
        def run_test(ix):
            lock.acquire()
            order.append(ix)
            lock.release()
        fibers = []
        for i in range(self.nfibers):
            fibers.append(gruvi.spawn(run_test, i))
        lock.acquire()
        gruvi.sleep(0)
        self.assertEqual(order, [])
        lock.release()
        for fib in fibers:
            fib.join()
        self.assertEqual(order, list(range(len(fibers))))

    def test_fiber_safety(self):
        # Start a bunch of fibers, each locking the rlock a few times before
        # unlocking it again. Ensure that the locks don't overlap.
        lock = self.Lock()
        failed = [0]
        def run_test():
            failed[0] += self.lock_unlock(lock, 20)
        fibers = []
        for i in range(self.nfibers):
            fiber = gruvi.Fiber(run_test)
            fiber.start()
            fibers.append(fiber)
        for fib in fibers:
            fib.join()
        self.assertEqual(failed[0], 0)

    def test_fiber_safety_timeout(self):
        # Test correctness of the lock in case of timeouts.
        lock = self.Lock()
        order = []
        def run_test(ix):
            lock.acquire()
            order.append(ix)
            lock.release()
        fibers = []
        for i in range(self.nfibers):
            fibers.append(gruvi.spawn(run_test, i))
        # There's 5 elements in lock._waiter now. Kill the first one, and
        # schedule a cancel for the second one. Both conditions should be
        # handled appropriately and no deadlocks should happen.
        lock.acquire()
        gruvi.sleep(0)  # make sure all fibers are waiting on lock.acquire()
        fibers[0].cancel()
        gruvi.sleep(0)  # first one will be gone
        self.assertFalse(fibers[0].is_alive())
        fibers[1].cancel()  # a Cancelled is now scheduled for number two
        self.assertTrue(fibers[1].is_alive())
        lock.release()
        for fib in fibers:
            fib.join()
        # All fibers should have gotten the lock except 1 and 2.
        self.assertEqual(order, list(range(2, len(fibers))))

    def test_thread_safety(self):
        # Start a bunch of threads, each starting a bunch of fibers. Each fiber
        # will lock the rlock a few times before unlocking it again. Ensure
        # that the locks don't overlap.
        lock = self.Lock()
        failed = []
        def run_test():
            failed.append(self.lock_unlock(lock, 10))
        def run_thread():
            fibers = []
            for i in range(self.nfibers):
                fiber = gruvi.Fiber(run_test)
                fiber.start()
                fibers.append(fiber)
            for fib in fibers:
                fib.join()
        threads = []
        for i in range(self.nthreads):
            thread = threading.Thread(target=run_thread)
            thread.start()
            threads.append(thread)
        for thread in threads:
            thread.join()
        self.assertEqual(sum(failed), 0)


class TestLock(UnitTest, LockTests):

    Lock = gruvi.Lock
    nfibers = 10
    nthreads = 2

    def lock_unlock(self, lock, count):
        failed = 0
        timeouts = 0
        for i in range(count):
            # the granularity of libuv's timers is 1ms.
            gruvi.sleep(random.randint(0, 2) / 1000)
            timeout = random.choice((None, 0.001))
            while not lock.acquire(timeout=timeout):
                timeouts += 1
            gruvi.sleep(random.randint(0, 2) / 1000)
            if lock._locked != 1 or lock._owner is not gruvi.current_fiber():
                failed += 1
            lock.release()
        return failed

    def test_timeout(self):
        # Ensure that the timeout argument to acquire() works.
        hub = get_hub()
        lock = gruvi.Lock()
        lock.acquire()
        t0 = hub.loop.now()
        self.assertFalse(lock.acquire(timeout=0.01))
        t1 = hub.loop.now()
        self.assertGreater(t1-t0, 10)
        self.assertFalse(lock._callbacks)

    def test_non_blocking(self):
        # Ensure that the blocking argument to acquire() works.
        lock = gruvi.Lock()
        lock.acquire()
        self.assertFalse(lock.acquire(blocking=False))
        self.assertFalse(lock._callbacks)

    def test_acquire_release_threads(self):
        # Ensure that a lock can be locked and unlocked in different threads.
        lock = self.Lock()
        sync = self.Lock()
        failed = [0]
        def thread_lock():
            lock.acquire()
            failed[0] += (1 if not lock.locked() else 0)
            sync.release()
        def thread_unlock():
            sync.acquire()
            lock.release()
            failed[0] += (1 if lock.locked() else 0)
            sync.release()
        sync.acquire()
        t1 = threading.Thread(target=thread_lock)
        t2 = threading.Thread(target=thread_unlock)
        t1.start(); t2.start()
        t1.join(); t2.join()
        self.assertEqual(failed[0], 0)


class TestRLock(UnitTest, LockTests):

    Lock = gruvi.RLock
    nfibers = 5
    nthreads = 2

    def lock_unlock(self, lock, count):
        failed = 0
        timeouts = 0
        for i in range(count):
            gruvi.sleep(random.randint(0, 2) / 1000)
            timeout = random.choice((None, 0.001))
            while not lock.acquire(timeout=timeout):
                timeouts += 1
            gruvi.sleep(random.randint(0, 2) / 1000)
            if lock._locked != 1 or lock._owner is not gruvi.current_fiber():
                failed += 1
            lock.acquire()
            gruvi.sleep(random.randint(0, 2) / 1000)
            if lock._locked != 2 or lock._owner is not gruvi.current_fiber():
                failed += 1
            lock.release()
            gruvi.sleep(random.randint(0, 2) / 1000)
            if lock._locked != 1 or lock._owner is not gruvi.current_fiber():
                failed += 1
            lock.release()
        return failed

    def test_multiple(self):
        # Lock and unlock the rlock a few times
        lock = gruvi.RLock()
        for i in range(5):
            lock.acquire()
            self.assertEqual(lock._locked, i+1)
        self.assertTrue(lock.locked())
        for i in range(5):
            lock.release()
            self.assertEqual(lock._locked, 4-i)
        self.assertFalse(lock.locked())

    def test_timeout(self):
        # Ensure that the timeout argument to acquire() works.
        hub = get_hub()
        lock = gruvi.RLock()
        sync = gruvi.Lock()
        def lock_rlock():
            lock.acquire()
            sync.acquire()
            lock.release()
        # This needs a new fiber, as the same fiber *can* lock the same RLock twice.
        sync.acquire()
        fiber = gruvi.spawn(lock_rlock)
        gruvi.sleep(0)
        self.assertTrue(lock.locked())
        t0 = hub.loop.now()
        self.assertFalse(lock.acquire(timeout=0.01))
        t1 = hub.loop.now()
        # Internally the event loop uses timestamps with a 1ms granularity. So
        # allow for that.
        self.assertGreaterEqual(t1-t0, 10)
        sync.release()
        fiber.join()
        self.assertFalse(lock._callbacks)

    def test_non_blocking(self):
        # Ensure that the blocking argument to acquire() works.
        lock = gruvi.RLock()
        sync = gruvi.Lock()
        def lock_rlock():
            lock.acquire()
            sync.acquire()
            lock.release()
        # This needs a new fiber, as the same fiber *can* lock the same RLock twice.
        sync.acquire()
        fiber = gruvi.spawn(lock_rlock)
        gruvi.sleep(0)
        self.assertTrue(lock.locked())
        self.assertFalse(lock.acquire(blocking=False))
        sync.release()
        fiber.join()
        self.assertFalse(lock._callbacks)


class TestEvent(UnitTest):

    nthreads = 10
    nfibers = 100

    def test_basic(self):
        # Ensure that an event can be set and cleared
        event = gruvi.Event()
        self.assertFalse(event.is_set())
        event.set()
        self.assertTrue(event.is_set())
        event.clear()
        self.assertFalse(event.is_set())

    def test_wait(self):
        event = gruvi.Event()
        done = []
        def waiter():
            done.append(False)
            done.append(event.wait())
        gruvi.spawn(waiter)
        gruvi.sleep(0)
        self.assertEqual(done, [False])
        event.set()
        gruvi.sleep(0)
        self.assertEqual(done, [False, True])

    def test_wait_timeout(self):
        event = gruvi.Event()
        self.assertFalse(event.wait(timeout=0.01))
        self.assertFalse(event._callbacks)

    def test_thread_safety(self):
        event = gruvi.Event()
        result = []
        def run_test():
            timeout = random.choice((None, 0.1))
            result.append(event.wait(timeout=timeout))
        def run_thread():
            fibers = []
            for i in range(self.nfibers):
                fibers.append(gruvi.spawn(run_test))
            for fib in fibers:
                fib.join()
        threads = []
        for i in range(self.nthreads):
            thread = threading.Thread(target=run_thread)
            thread.start()
            threads.append(thread)
        gruvi.sleep(0.5)
        event.set()
        for thread in threads:
            thread.join()
        timeouts = result.count(False)
        notified = result.count(True)
        self.assertEqual(timeouts + notified, self.nthreads*self.nfibers)
        self.assertGreater(timeouts, 0)
        self.assertGreater(notified, 0)
        self.assertFalse(event._callbacks)


class TestCondition(UnitTest):

    nthreads = 5
    nfibers = 100

    def test_basic(self):
        # Ensure that a basic wait/notify works.
        cond = gruvi.Condition()
        waiting = [0]
        def wait_cond():
            with cond:
                waiting[0] += 1
                cond.wait()
                waiting[0] -= 1
        gruvi.spawn(wait_cond)
        gruvi.sleep(0)
        self.assertEqual(waiting[0], 1)
        with cond:
            cond.notify()
        gruvi.sleep(0)
        self.assertEqual(waiting[0], 0)

    def test_notify_multiple(self):
        # Ensure that multiple fibers can be notified, and that the order in
        # which they are notified is respected.
        cond = gruvi.Condition()
        waiting = [0]
        done = []
        def wait_cond(i):
            with cond:
                waiting[0] += 1
                cond.wait()
                waiting[0] -= 1
                done.append(i)
        fibers = []
        for i in range(10):
            fibers.append(gruvi.spawn(wait_cond, i))
        gruvi.sleep(0)
        self.assertEqual(waiting[0], 10)
        with cond:
            cond.notify(1)
        gruvi.sleep(0)
        self.assertEqual(waiting[0], 9)
        with cond:
            cond.notify(3)
        gruvi.sleep(0)
        self.assertEqual(waiting[0], 6)
        with cond:
            cond.notify_all()
        gruvi.sleep(0)
        self.assertEqual(waiting[0], 0)
        self.assertEqual(done, list(range(10)))

    def test_wait_for(self):
        # Ensure that wait_for can wait for a predicate
        cond = gruvi.Condition()
        waiting = [0]
        unblock = []
        done = []
        def wait_cond(i):
            with cond:
                waiting[0] += 1
                cond.wait_for(lambda: i in unblock)
                waiting[0] -= 1
                done.append(i)
        fibers = []
        for i in range(10):
            fibers.append(gruvi.spawn(wait_cond, i))
        gruvi.sleep(0)
        self.assertEqual(waiting[0], 10)
        with cond:
            cond.notify(1)  # no predicate matches
        gruvi.sleep(0)
        self.assertEqual(waiting[0], 10)
        unblock += [0]
        with cond:
            cond.notify(1)  # one predicate matches
        gruvi.sleep(0)
        self.assertEqual(waiting[0], 9)
        unblock += [2, 3]
        with cond:
            cond.notify(3)  # two match
        gruvi.sleep(0)
        self.assertEqual(waiting[0], 7)
        unblock += [1]
        with cond:
            cond.notify_all()  # one match
        gruvi.sleep(0)
        self.assertEqual(waiting[0], 6)
        unblock += list(range(10))
        with cond:
            cond.notify_all()  # one match
        gruvi.sleep(0)
        self.assertEqual(waiting[0], 0)
        self.assertEqual(done, [0, 2, 3, 1, 4, 5, 6, 7, 8, 9])

    def test_call_without_lock(self):
        # A RuntimeError should be raised if notify or wait are called without
        # the lock.
        cond = gruvi.Condition()
        self.assertRaises(RuntimeError, cond.wait)
        self.assertRaises(RuntimeError, cond.notify)

    def test_wait_timeout(self):
        # When a timeout occurs, wait() should return False
        cond = gruvi.Condition()
        with cond:
            self.assertFalse(cond.wait(timeout=0.01))
        self.assertFalse(cond._callbacks)

    def test_wait_for_timeout(self):
        # When a timeout occurs, wait_for() should return False
        cond = gruvi.Condition()
        waiters = [0]
        def notify_cond():
            with cond:
                waiters[0] += 1
                cond.notify()
                waiters[0] -= 1
        gruvi.spawn(notify_cond)
        with cond:
            self.assertEqual(waiters[0], 0)
            self.assertFalse(cond.wait_for(lambda: False, timeout=0.1))
            self.assertEqual(waiters[0], 0)

    def test_thread_safety(self):
        cond = gruvi.Condition()
        ready = []
        result = []
        def run_test():
            with cond:
                ready.append(gruvi.current_fiber())
                timeout = random.choice((None, 0.1))
                result.append(cond.wait(timeout=timeout))
        def run_thread():
            fibers = []
            for i in range(self.nfibers):
                fibers.append(gruvi.spawn(run_test))
            for fib in fibers:
                fib.join()
        threads = []
        for i in range(self.nthreads):
            thread = threading.Thread(target=run_thread)
            thread.start()
            threads.append(thread)
        gruvi.sleep(0.2)
        while len(ready) != self.nthreads*self.nfibers:
            gruvi.sleep(0.1)
        with cond:
            cond.notify_all()
        for thread in threads:
            thread.join()
        timeouts = result.count(False)
        notified = result.count(True)
        self.assertEqual(timeouts + notified, self.nthreads*self.nfibers)
        self.assertGreater(timeouts, 0)
        self.assertGreater(notified, 0)
        self.assertFalse(cond._callbacks)


class TestQueue(UnitTest):

    def test_basic(self):
        # What is put in the queue, should come out.
        queue = gruvi.Queue()
        queue.put(10)
        self.assertEqual(queue.get(), 10)

    def test_types(self):
        # Queue should support putting in arbitrary objects.
        queue = gruvi.Queue()
        queue.put('foo')
        self.assertEqual(queue.get(), 'foo')
        queue.put(['foo'])
        self.assertEqual(queue.get(), ['foo'])

    def test_order(self):
        # The behavior of a queue should be FIFO
        queue = gruvi.Queue()
        for i in range(10):
            queue.put(10+i)
        for i in range(10):
            self.assertEqual(queue.get(), 10+i)

    def test_qsize(self):
        # The qsize() of a queue should by default be the number of elements
        queue = gruvi.Queue()
        for i in range(10):
            queue.put(10+i)
            self.assertEqual(queue.qsize(), i+1)
        for i in range(10):
            self.assertEqual(queue.get(), 10+i)
            self.assertEqual(queue.qsize(), 10-i-1)

    def test_qsize_custom_size(self):
        # The put() method has an optional "size" argument that allows you to
        # specify a custom size.
        queue = gruvi.Queue()
        for i in range(10):
            queue.put(10+i, size=2)
            self.assertEqual(queue.qsize(), 2*(i+1))
        for i in range(10):
            self.assertEqual(queue.get(), 10+i)
            self.assertEqual(queue.qsize(), 2*(10-i-1))

    def test_get_wait(self):
        # Queue.get() should wait until an item becomes available.
        queue = gruvi.Queue()
        def put_queue(value):
            gruvi.sleep(0.01)
            queue.put(value)
        gruvi.spawn(put_queue, 'foo')
        self.assertEqual(queue.get(), 'foo')

    def test_get_timeout(self):
        # Ensure the "timeout" argument to Queue.get() works
        queue = gruvi.Queue()
        hub = get_hub()
        t0 = hub.loop.now()
        self.assertRaises(gruvi.QueueEmpty, queue.get, timeout=0.01)
        t1 = hub.loop.now()
        self.assertGreaterEqual(t1-t0, 10)

    def test_get_non_blocking(self):
        # Ensure the "block" argument to Queue.get() works
        queue = gruvi.Queue()
        self.assertRaises(gruvi.QueueEmpty, queue.get, block=False)
        self.assertRaises(gruvi.QueueEmpty, queue.get_nowait)

    def test_put_timeout(self):
        # Ensure the "timeout" argument to Queue.put() works
        queue = gruvi.Queue(maxsize=10)
        queue.put('foo', size=10)
        hub = get_hub()
        t0 = hub.loop.now()
        self.assertRaises(gruvi.QueueFull, queue.put, 'bar', timeout=0.01)
        t1 = hub.loop.now()
        self.assertGreaterEqual(t1-t0, 10)

    def test_put_non_blocking(self):
        # Ensure the "block" argument to Queue.put() works
        queue = gruvi.Queue(maxsize=10)
        queue.put('foo', size=10)
        self.assertRaises(gruvi.QueueFull, queue.put, 'bar', block=False)
        self.assertRaises(gruvi.QueueFull, queue.put_nowait, 'bar')

    def test_empty(self):
        # Ensure that empty() returns nonzero if the queue is empty.
        queue = gruvi.Queue()
        self.assertTrue(queue.empty())
        queue.put('foo')
        self.assertFalse(queue.empty())

    def test_full(self):
        # Ensure that empty() returns nonzero if the queue is empty.
        queue = gruvi.Queue(maxsize=1)
        self.assertFalse(queue.full())
        queue.put('foo')
        self.assertTrue(queue.full())

    def test_task_done(self):
        # Ensure that task_done properly updates the unifished_tasks property.
        queue = gruvi.Queue()
        self.assertEqual(queue.unfinished_tasks, 0)
        queue.put('foo')
        self.assertEqual(queue.unfinished_tasks, 1)
        self.assertEqual(queue.get(), 'foo')
        self.assertEqual(queue.unfinished_tasks, 1)
        queue.task_done()
        self.assertEqual(queue.unfinished_tasks, 0)
        self.assertRaises(RuntimeError, queue.task_done)
        self.assertEqual(queue.unfinished_tasks, 0)

    def test_join_wait(self):
        # Ensure that join() waits until all unfished tasks are done.
        queue = gruvi.Queue()
        self.assertEqual(queue.unfinished_tasks, 0)
        queue.put('foo')
        def consumer():
            queue.get()
            gruvi.sleep(0.01)
            queue.task_done()
        gruvi.spawn(consumer)
        self.assertEqual(queue.unfinished_tasks, 1)
        queue.join()
        self.assertEqual(queue.unfinished_tasks, 0)

    def test_produce_consume(self):
        # Ensure that there's no deadlocks when pushing a large number of items
        # through a queue with a fixed size.
        queue = gruvi.Queue(maxsize=10)
        result = []; sizes = []
        def consumer(n):
            for i in range(n):
                queue.put(i)
                sizes.append(queue.qsize())
        def producer(n):
            for i in range(n):
                result.append(queue.get())
                sizes.append(queue.qsize())
        ni = 2000
        fcons = gruvi.spawn(consumer, ni)
        fprod = gruvi.spawn(producer, ni)
        fcons.join(); fprod.join()
        self.assertEqual(len(result), ni)
        self.assertEqual(result, list(range(ni)))
        self.assertLessEqual(max(sizes), 10)

    def test_thread_safety(self):
        # A Queue should be thread safe. This meanst that all entries that are
        # put in the queue must be returned, that no entry must be returned
        # twice and that the order must be respected. Also no deadlock must
        # ever occur.
        # To test, fire up a bunch of threads which each fire up a bunch of
        # fibers, and have the fibers do some random sleeps. Then let it run
        # and test the result.
        result = []
        reference = []
        lock = gruvi.Lock()
        def put_queue(tid, fid, count):
            for i in range(count):
                with lock:
                    gruvi.sleep(random.randint(0, 2) / 1000)
                    queue.put((tid, fid, count))
                    reference.append((tid, fid, count))
        def get_queue(count):
            for i in range(count):
                with lock:
                    result.append(queue.get())
        def thread_put(tid, nfibers, count):
            fibers = []
            for i in range(nfibers):
                fibers.append(gruvi.spawn(put_queue, tid, i, count))
            for fib in fibers:
                fib.join()
            gruvi.get_hub().close()
        def thread_get(nfibers, count):
            fibers = []
            for i in range(nfibers):
                fibers.append(gruvi.spawn(get_queue, count))
            for fib in fibers:
                fib.join()
            gruvi.get_hub().close()
        queue = gruvi.Queue()
        threads = []
        # 5 procuders and 5 consumers, each with 20 fibers
        for i in range(5):
            thread = threading.Thread(target=thread_put, args=(i, 20, 5))
            thread.start()
            threads.append(thread)
        for i in range(5):
            thread = threading.Thread(target=thread_get, args=(20, 5))
            thread.start()
            threads.append(thread)
        for thread in threads:
            thread.join()
        gruvi.sleep(0)  # run callbacks
        self.assertEqual(len(result), 500)
        self.assertEqual(result, reference)
        # Within a (tid,fid) pair, the counts must be monotonic
        partial_sort = sorted(result, key=lambda el: (el[0], el[1]))
        full_sort = sorted(result, key=lambda el: (el[0], el[1], el[2]))
        self.assertEqual(partial_sort, full_sort)


class TestLifoQueue(UnitTest):

    def test_order(self):
        # The behavior of a queue should be LIFO
        queue = gruvi.LifoQueue()
        for i in range(10):
            queue.put(10+i)
        for i in range(10):
            self.assertEqual(queue.get(), 19-i)


class TestPriorityQueue(UnitTest):

    def test_order(self):
        # The queue should respect the priority we give it.
        queue = gruvi.PriorityQueue()
        items = list(range(100))
        prios = list(range(100))
        random.shuffle(prios)
        items = list(zip(prios, items))
        for item in items:
            queue.put(item)
        result = []
        for i in range(len(items)):
            result.append(queue.get())
        self.assertEqual(sorted(items), result)


if __name__ == '__main__':
    unittest.main()



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import os
import sys

from argparse import ArgumentParser
from unittest import TestLoader, TextTestRunner, TestSuite


parser = ArgumentParser()
parser.add_argument('-v', '--verbose', help='be more verbose', action='count', default=1)
parser.add_argument('-f', '--failfast', help='stop on first failure', action='store_true')
parser.add_argument('-b', '--buffer', help='buffer stdout and stderr', action='store_true')
parser.add_argument('suite', nargs='+', help='name of test suite to run', metavar='suite',
                    choices=('all', 'unit', 'performance', 'memory', 'examples'))
args = parser.parse_args()

if 'all' in args.suite:
    args.suite = ['unit', 'performance', 'memory', 'examples']

# $VERBOSE can override -v
try:
    verbose = int(os.environ['VERBOSE'])
except (KeyError, ValueError):
    verbose = args.verbose
os.environ['VERBOSE'] = str(verbose)

# Change directory to tests/ irrespective of where we're called from.
topdir = os.path.split(os.path.abspath(__file__))[0]
testdir = os.path.join(topdir, 'tests')
os.chdir(testdir)

sys.path.insert(0, testdir)
from support import TestCase, MemoryTest, PerformanceTest

suite = TestSuite()

for name in args.suite:
    TestCase.setup_loader()
    if name == 'unit':
        pattern = 'test_*.py'
    elif name == 'performance':
        pattern = 'perf_*.py'
        PerformanceTest.setup_loader()
        PerformanceTest.start_new_results()
    elif name == 'memory':
        pattern = 'memory.py'
        MemoryTest.setup_loader()
        MemoryTest.start_new_results()
    elif name == 'examples':
        pattern = 'examples.py'
    loader = TestLoader()
    tests = loader.discover('.', pattern)
    suite.addTest(tests)

runner = TextTestRunner(verbosity=verbose, buffer=args.buffer, failfast=args.failfast)
result = runner.run(suite)
if result.errors or result.failures:
    sys.exit(1)



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2017 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import sys
import re
import functools

from weakref import WeakKeyDictionary

__all__ = []


class AbsentType(object):
    """A type that represents the absence of a value.

    Useful in parsing protocols where there's a difference between a NULL
    value, and an absent value.
    """

    def __nonzero__(self):
        return False

    __bool__ = __nonzero__

    def __repr__(self):
        return 'Absent'

    __str__ = __repr__

Absent = AbsentType()


def docfrom(base):
    """Decorator to set a function's docstring from another function."""
    def setdoc(func):
        func.__doc__ = (getattr(base, '__doc__') or '') + (func.__doc__ or '')
        return func
    return setdoc


_objrefs = WeakKeyDictionary()  # obj -> objref
_lastids = {}  # classname -> lastid

def objref(obj):
    """Return a string that uniquely and compactly identifies an object."""
    ref = _objrefs.get(obj)
    if ref is None:
        clsname = obj.__class__.__name__.split('.')[-1]
        seqno = _lastids.setdefault(clsname, 1)
        ref = '{}-{}'.format(clsname, seqno)
        _objrefs[obj] = ref
        _lastids[clsname] += 1
    return ref


re_lu = re.compile('[A-Z]+[a-z0-9]+')

def split_cap_words(s):
    """Split the CamelCase string *s* into words."""
    return re_lu.findall(s)


def delegate_method(other, method, name=None):
    """Add a method to the current class that delegates to another method.

    The *other* argument must be a property that returns the instance to
    delegate to. Due to an implementation detail, the property must be defined
    in the current class. The *method* argument specifies a method to delegate
    to. It can be any callable as long as it takes the instances as its first
    argument.

    It is a common paradigm in Gruvi to expose protocol methods onto clients.
    This keeps most of the logic into the protocol, but prevents the user from
    having to type ``'client.protocol.*methodname*'`` all the time.

    For example::

      class MyClient(Client):

          protocol = Client.protocol

          delegate_method(protocol, MyProtocol.method)
    """
    frame = sys._getframe(1)
    classdict = frame.f_locals

    @functools.wraps(method)
    def delegate(self, *args, **kwargs):
        other_self = other.__get__(self)
        return method(other_self, *args, **kwargs)

    if getattr(method, '__switchpoint__', False):
        delegate.__switchpoint__ = True

    if name is None:
        name = method.__name__
    propname = None
    for key in classdict:
        if classdict[key] is other:
            propname = key
            break
    # If we know the property name, replace the docstring with a small
    # reference instead of copying the function docstring.
    if propname:
        qname = getattr(method, '__qualname__', method.__name__)
        if '.' in qname:
            delegate.__doc__ = 'A shorthand for ``self.{propname}.{name}()``.' \
                               .format(name=name, propname=propname)
        else:
            delegate.__doc__ = 'A shorthand for ``{name}({propname}, ...)``.' \
                               .format(name=name, propname=propname)
    classdict[name] = delegate



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import time
import unittest

import gruvi
from gruvi.fibers import Fiber
from support import PerformanceTest


class PerfFiber(PerformanceTest):

    def perf_spawn_throughput(self):
        # Measure the number of fibers we can spawn per second.
        t0 = t1 = time.time()
        count = [0]
        def dummy_fiber():
            count[0] += 1
        while t1 - t0 < 0.2:
            fiber = Fiber(dummy_fiber)
            fiber.start()
            gruvi.sleep(0)
            t1 = time.time()
        speed = count[0] / (t1 - t0)
        self.add_result(speed)

    def perf_switch_throughput(self):
        # Measure the number of switches we can do per second.
        t0 = t1 = time.time()
        count = [0]
        def switch_parent():
            while True:
                gruvi.sleep(0)
        fiber = Fiber(switch_parent)
        fiber.start()
        while t1 - t0 < 0.2:
            gruvi.sleep(0)
            count[0] += 1
            t1 = time.time()
        speed = count[0] / (t1 - t0)
        self.add_result(speed)
        fiber.cancel()
        gruvi.sleep(0)


if __name__ == '__main__':
    unittest.defaultTestLoader.testMethodPrefix = 'perf'
    unittest.main()



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function, division

import os
import time
import unittest
from unittest import SkipTest

from gruvi.ssl import SslPipe
from support import PerformanceTest
from test_ssl import communicate


class PerfSsl(PerformanceTest):

    def setUp(self):
        if not os.access(self.certname, os.R_OK):
            raise SkipTest('no certificate available')
        super(PerfSsl, self).setUp()
        context = self.get_ssl_context()
        self.client = SslPipe(context, False)
        self.server = SslPipe(context, True)

    def perf_throughput(self):
        client, server = self.client, self.server
        buf = b'x' * 65536
        nbytes = 0
        clientssl = client.do_handshake()
        server.do_handshake()
        t0 = t1 = time.time()
        while t1 - t0 < 0.2:
            received = communicate(buf, client, server, clientssl, [])
            if clientssl:
                clientssl = []
            nbytes += len(received)
            t1 = time.time()
        cipher = server.ssl.cipher()[0].lower().replace('-', '_')
        name = 'ssl_throughput_{0}'.format(cipher)
        speed = nbytes / (t1 - t0) / (1024 * 1024)
        self.add_result(speed, name=name)


if __name__ == '__main__':
    unittest.defaultTestLoader.testMethodPrefix = 'perf'
    unittest.main()



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

from .error import *
from .message import (DBusMessage, MethodCallMessage, MethodReturnMessage,
                      ErrorMessage, SignalMessage, parseMessage)
from .authentication import ClientAuthenticator, BusAuthenticator

# A few renames to have a consistent naming scheme
DbusMessage = DBusMessage
DbusException = DBusException
DbusAuthenticationFailed = DBusAuthenticationFailed



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import pyuv
from .callbacks import add_callback, remove_callback, walk_callbacks
from .callbacks import has_callback, clear_callbacks

__all__ = ['READABLE', 'WRITABLE']


READABLE = pyuv.UV_READABLE
WRITABLE = pyuv.UV_WRITABLE


if __debug__:

    def dump(mp):
        print('== Dumping MultiPoll {!r}'.format(mp))
        # [total, nreaders, nwriters, nreadwrite, disabled]
        stats = [0, 0, 0, 0, 0]
        def walker(callback, args):
            stats[0] += 1
            if args == READABLE:
                stats[1] += 1
            elif args == WRITABLE:
                stats[2] += 1
            elif args == READABLE|WRITABLE:
                stats[3] += 1
            elif not args:
                stats[4] += 1
            return True
        walk_callbacks(mp, walker)
        print('Using Poll: {!r}'.format(mp._poll))
        print('Poll FD: {}'.format(mp._poll.fileno()))
        print('Total callbacks: {}'.format(stats[0]))
        print('Callbacks registered for read: {}'.format(stats[1]))
        print('Callbacks registered for write: {}'.format(stats[2]))
        print('Callbacks registered for read|write: {}'.format(stats[3]))
        print('Callbacks that are inactive: {}'.format(stats[4]))

    def check(mp):
        counts = [0, 0]  # [readable, writable]
        errors = [0]
        def walker(callback, events):
            if events & READABLE:
                counts[0] += 1
            if events & WRITABLE:
                counts[1] += 1
            if events & ~(READABLE|WRITABLE):
                errors[0] += 1
            return True
        walk_callbacks(mp, walker)
        assert mp._readers == counts[0]
        assert mp._writers == counts[1]
        assert errors[0] == 0
        events = (READABLE if counts[0] > 0 else 0) \
                        | (WRITABLE if counts[1] > 0 else 0)
        assert mp._events == events


class MultiPoll(object):
    """A version of :class:`pyuv.Poll` that allows connecting multiple readers
    and writers to the same file descriptor.
    """

    __slots__ = ('_poll', '_readers', '_writers', '_callbacks', '_events')

    def __init__(self, loop, fd):
        self._poll = pyuv.Poll(loop, fd)
        self._readers = 0
        self._writers = 0
        self._callbacks = None
        self._events = 0

    def _run_callbacks(self, handle, events, error):
        # Run those callbacks that were registered for *events*.
        if error:
            return
        def run_callback(callback, args):
            masked = args & events
            if masked:
                callback(self._poll.fileno(), masked)
            return True  # Keep callback, poll callbacks are persistent
        walk_callbacks(self, run_callback)

    def _sync(self):
        # Synchronize the current set of events with the pyuv.Poll object.
        events = (READABLE if self._readers > 0 else 0) \
                        | (WRITABLE if self._writers > 0 else 0)
        if events == self._events:
            return
        if self._events:
            self._poll.stop()
        if events:
            self._poll.start(events, self._run_callbacks)
        self._events = events

    def add_callback(self, events, callback):
        """Add a new callback."""
        if self._poll is None:
            raise RuntimeError('poll instance is closed')
        if events & ~(READABLE|WRITABLE):
            raise ValueError('illegal event mask: {}'.format(events))
        if events & READABLE:
            self._readers += 1
        if events & WRITABLE:
            self._writers += 1
        handle = add_callback(self, callback, events)
        self._sync()
        return handle

    def remove_callback(self, handle):
        """Remove a callback."""
        if self._poll is None:
            raise RuntimeError('poll instance is closed')
        remove_callback(self, handle)
        if handle.args & READABLE:
            self._readers -= 1
        if handle.args & WRITABLE:
            self._writers -= 1
        self._sync()

    def update_callback(self, handle, events):
        """Update the event mask for a callback."""
        if self._poll is None:
            raise RuntimeError('poll instance is closed')
        if not has_callback(self, handle):
            raise ValueError('no such callback')
        if events & ~(READABLE|WRITABLE):
            raise ValueError('illegal event mask: {}'.format(events))
        if handle.args == events:
            return
        if handle.args & READABLE:
            self._readers -= 1
        if handle.args & WRITABLE:
            self._writers -= 1
        if events & READABLE:
            self._readers += 1
        if events & WRITABLE:
            self._writers += 1
        handle.args = events
        self._sync()

    def close(self):
        """Close the poll instance."""
        if self._poll is None:
            return
        self._poll.close()
        self._poll = None
        self._readers = 0
        self._writers = 0
        self._events = 0
        clear_callbacks(self)


class Poller(object):
    """A flexible file descriptor watcher.

    A Poller can watch multiple file descriptors, and each file descriptors can
    have multiple callbacks registered to it.

    This object is useful when integrating with third party libraries that
    support asynchronous IO by exposing a file descriptor that can be added to
    the event loop.

    Normally you should not instantiate this class yourself. Instead, use the
    per-loop instance that is available as :attr:`Hub.poller`.
    """

    def __init__(self, loop):
        self._loop = loop
        self._mpoll = {}

    def add_callback(self, fd, events, callback):
        """Add a new callback.

        The file descriptor *fd* will be watched for the events specified by
        the *events* parameter, which should be a bitwise OR of the constants
        ``READABLE`` and ``WRITABLE``. Whenever one or more of the specified
        events occur, *callback* will be called with a single integer argument
        containing the bitwise OR of the current events.

        The return value of this method is an opaque handle that can be used
        to remove the callback or to update the events associated with it.
        """
        if self._mpoll is None:
            raise RuntimeError('Poller instance is closed')
        try:
            mpoll = self._mpoll[fd]
        except KeyError:
            mpoll = self._mpoll[fd] = MultiPoll(self._loop, fd)
        handle = mpoll.add_callback(events, callback)
        return handle

    def remove_callback(self, fd, handle):
        """Remove a callback added by :meth:`~Poller.add_callback`.

        If this is the last callback that is registered for the fd then this
        will deallocate the ``MultiPoll`` instance and close the libuv handle.
        """
        if self._mpoll is None:
            raise RuntimeError('Poller instance is closed')
        mpoll = self._mpoll.get(fd)
        if mpoll is None:
            raise ValueError('not watching fd {}'.format(fd))
        mpoll.remove_callback(handle)

    def update_callback(self, fd, handle, events):
        """Update the event mask associated with an existing callback.

        If you want to temporarily disable a callback then you can use this
        method with an *events* argument of ``0``. This is more efficient than
        removing the callback and adding it again later.
        """
        if self._mpoll is None:
            raise RuntimeError('Poller instance is closed')
        mpoll = self._mpoll.get(fd)
        if mpoll is None:
            raise ValueError('not watching fd {}'.format(fd))
        mpoll.update_callback(handle, events)

    def close(self):
        """Close all active poll instances and remove all callbacks."""
        if self._mpoll is None:
            return
        for mpoll in self._mpoll.values():
            mpoll.close()
        self._mpoll.clear()
        self._mpoll = None



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import socket
import unittest

import gruvi
from gruvi.stream import StreamProtocol
from gruvi.endpoints import create_server, create_connection, getaddrinfo
from gruvi.transports import TransportError

from support import UnitTest


class TestCreateConnection(UnitTest):

    def test_tcp(self):
        # Ensure that create_connection() and create_server() can be used to
        # connect to each other over TCP.
        server = create_server(StreamProtocol, ('localhost', 0))
        addr = server.addresses[0]
        ctrans, cproto = create_connection(StreamProtocol, addr)
        gruvi.sleep(0.1)  # allow Server to accept()
        strans, sproto = list(server.connections)[0]
        cproto.stream.write(b'foo\n')
        self.assertEqual(sproto.stream.readline(), b'foo\n')
        server.close()
        self.assertEqual(len(list(server.connections)), 0)
        self.assertEqual(cproto.stream.readline(), b'')
        ctrans.close()

    def test_pipe(self):
        # Ensure that create_connection() and create_server() can be used to
        # connect to each other over a pipe.
        addr = self.pipename()
        server = create_server(StreamProtocol, addr)
        ctrans, cproto = create_connection(StreamProtocol, addr)
        gruvi.sleep(0.1)  # allow Server to accept()
        strans, sproto = list(server.connections)[0]
        cproto.stream.write(b'foo\n')
        self.assertEqual(sproto.stream.readline(), b'foo\n')
        server.close()
        self.assertEqual(len(list(server.connections)), 0)
        self.assertEqual(cproto.stream.readline(), b'')
        ctrans.close()

    def test_tcp_ssl(self):
        # Ensure that create_connection() and create_server() can be used to
        # connect to each other over TCP using SSL.
        context = self.get_ssl_context()
        server = create_server(StreamProtocol, ('localhost', 0), ssl=context)
        addr = server.addresses[0]
        ctrans, cproto = create_connection(StreamProtocol, addr, ssl=context)
        # No need to sleep here because create_connection waits for the SSL
        # handshake to complete
        strans, sproto = list(server.connections)[0]
        cproto.stream.write(b'foo\n')
        self.assertEqual(sproto.stream.readline(), b'foo\n')
        server.close()
        self.assertEqual(len(list(server.connections)), 0)
        self.assertEqual(cproto.stream.readline(), b'')
        ctrans.close()

    def test_pipe_ssl(self):
        # Ensure that create_connection() and create_server() can be used to
        # connect to each other over a pipe using SSL.
        context = self.get_ssl_context()
        addr = self.pipename()
        server = create_server(StreamProtocol, addr, ssl=context)
        ctrans, cproto = create_connection(StreamProtocol, addr, ssl=context)
        # No need to sleep here because create_connection waits for the SSL
        # handshake to complete
        strans, sproto = list(server.connections)[0]
        cproto.stream.write(b'foo\n')
        self.assertEqual(sproto.stream.readline(), b'foo\n')
        server.close()
        self.assertEqual(len(list(server.connections)), 0)
        self.assertEqual(cproto.stream.readline(), b'')
        ctrans.close()

    def test_ssl_handshake_on_connect(self):
        # Ensure that when create_connection(ssl=True) returns, that the SSL
        # handshake has completed.
        context = self.get_ssl_context()
        server = create_server(StreamProtocol, ('localhost', 0), ssl=context)
        addr = server.addresses[0]
        ctrans, cproto = create_connection(StreamProtocol, addr, ssl=context)
        # The SSL handshake should be completed at this point. This means that
        # there should be a channel binding.
        sslobj = ctrans.get_extra_info('ssl')
        self.assertIsNotNone(sslobj)
        sslcb = sslobj.get_channel_binding('tls-unique')
        self.assertGreater(len(sslcb), 0)
        strans, sproto = list(server.connections)[0]
        cproto.stream.write(b'foo\n')
        self.assertEqual(sproto.stream.readline(), b'foo\n')
        server.close()
        self.assertEqual(len(list(server.connections)), 0)
        self.assertEqual(cproto.stream.readline(), b'')
        ctrans.close()

    def test_ssl_no_handshake_on_connect(self):
        # Ensure that when create_connection(ssl=True) returns, that the SSL
        # handshake has completed.
        context = self.get_ssl_context()
        server = create_server(StreamProtocol, ('localhost', 0), ssl=context)
        addr = server.addresses[0]
        ssl_args = {'do_handshake_on_connect': False}
        ctrans, cproto = create_connection(StreamProtocol, addr, ssl=context, ssl_args=ssl_args)
        # The SSL handshake has not been established at this point.
        sslobj = ctrans.get_extra_info('ssl')
        self.assertIsNone(sslobj)
        # Now initiate the SSL handshake and allow it some time to complete
        ctrans.do_handshake()
        gruvi.sleep(0.1)
        # There should be a channel binding now.
        sslobj = ctrans.get_extra_info('ssl')
        self.assertIsNotNone(sslobj)
        sslcb = sslobj.get_channel_binding('tls-unique')
        self.assertGreater(len(sslcb), 0)
        strans, sproto = list(server.connections)[0]
        cproto.stream.write(b'foo\n')
        self.assertEqual(sproto.stream.readline(), b'foo\n')
        server.close()
        self.assertEqual(len(list(server.connections)), 0)
        self.assertEqual(cproto.stream.readline(), b'')
        ctrans.close()

    def test_tcp_connection_refused(self):
        # Ensure that create_connection raises an exception when connecting to
        # a TCP port where there is no listener.
        server = create_server(StreamProtocol, ('localhost', 0))
        addr = server.addresses[0]
        server.close()
        self.assertRaises(TransportError, create_connection, StreamProtocol, addr)

    def test_pipe_connection_refused(self):
        # Ensure that create_connection raises an exception when connecting to
        # a Pipe that has no listener.
        addr = self.pipename()
        self.assertRaises(TransportError, create_connection, StreamProtocol, addr)


class TestGetAddrInfo(UnitTest):

    def test_resolve(self):
        res = getaddrinfo('localhost', family=socket.AF_INET)
        self.assertIsInstance(res, list)
        self.assertGreater(len(res), 0)
        self.assertEqual(res[0].sockaddr, ('127.0.0.1', 0))

    def get_resolve_timeout(self):
        self.assertRaises(gruvi.Timeout, getaddrinfo, 'localhost', timeout=0)


if __name__ == '__main__':
    unittest.main()



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import sys
import socket
import _ssl
import ssl
import six

__all__ = []

if sys.version_info[:2] < (3, 5):
    from . import _sslcompat

    # Export constants from _sslcompat
    for key in dir(_sslcompat):
        if key[:4].isupper():  # also export FOO_BARv1
            value = getattr(_sslcompat, key)
            globals()[key] = value
    errorcode = _sslcompat.errorcode

if hasattr(ssl, 'CHANNEL_BINDING_TYPES'):
    CHANNEL_BINDING_TYPES = ssl.CHANNEL_BINDING_TYPES
elif _sslcompat.HAS_TLS_UNIQUE:
    CHANNEL_BINDING_TYPES = ['tls-unique']
else:
    CHANNEL_BINDING_TYPES = []

if hasattr(ssl, 'MemoryBIO'):
    MemoryBIO = ssl.MemoryBIO
else:
    MemoryBIO = _sslcompat.MemoryBIO


if hasattr(ssl, 'SSLContext'):

    # Python 2.7.9 and Python 3.x

    SSLContext = ssl.SSLContext

else:

    # Python 2.7.x, x <= 8

    class SSLContext(object):

        # This isn't a real SSLContext so it doesn't do things like session
        # caching. The purpose is to store arguments to :func:`ssl.wrap_socket`
        # in a way that is compatible with Py3k.

        def __init__(self, protocol):
            # [keyfile, certfile, cert_reqs, ssl_version, ca_certs, ciphers]
            self._ssl_args = [None, None, ssl.CERT_NONE, protocol, None, ssl._DEFAULT_CIPHERS]
            # Implement the same defaults as the Python ssl module
            self._options = _sslcompat.OP_ALL & _sslcompat.OP_NO_SSLv2
            self._dh_params = None

        @property
        def protocol(self):
            return self._ssl_args[3]

        @property
        def verify_mode(self):
            return self._ssl_args[2]

        @verify_mode.setter
        def verify_mode(self, cert_reqs):
            self._ssl_args[2] = cert_reqs

        @property
        def options(self):
            return self._options

        @options.setter
        def options(self, options):
            self._options = options

        def load_verify_locations(self, ca_certs):
            self._ssl_args[4] = ca_certs

        def load_cert_chain(self, certfile, keyfile):
            self._ssl_args[0] = keyfile
            self._ssl_args[1] = certfile

        def set_ciphers(self, ciphers):
            self._ssl_args[5] = ciphers

        def load_dh_params(self, dh_params):
            self._dh_params = dh_params


if hasattr('ssl', 'SSLObject'):

    # Python 3.5

    SSLObject = ssl.SSLObject

else:

    # Python 2.7, 3.3 and 3.4
    # SSLObject was copied and adapted from Python 3.5

    class SSLObject(object):
        """An :class:`ssl.SSLObject` implementation for Python 2.7, 3.3 and 3.4."""

        def __init__(self, sslobj, sslctx, server_side, server_hostname):
            self._sslobj = sslobj
            self._sslctx = sslctx
            self._server_side = server_side
            self._server_hostname = server_hostname

        @property
        def context(self):
            return self._sslctx

        @context.setter
        def context(self, ctx):
            if not hasattr(self._sslobj, 'context'):
                raise RuntimeError('cannot set context on Python 2.7')
            self._sslobj.context = ctx
            self._sslctx = ctx

        @property
        def server_side(self):
            return self._server_side

        @property
        def server_hostname(self):
            return self._server_hostname

        def read(self, len=0, buffer=None):
            if buffer is not None:
                v = self._sslobj.read(len, buffer)
            else:
                v = self._sslobj.read(len or 1024)
            return v

        def write(self, data):
            return self._sslobj.write(data)

        def getpeercert(self, binary_form=False):
            return self._sslobj.peer_certificate(binary_form)

        def selected_npn_protocol(self):
            if hasattr(self._sslobj, 'selected_npn_protocol'):
                return self._sslobj.selected_npn_protocol()

        def cipher(self):
            return self._sslobj.cipher()

        def compression(self):
            if hasattr(self._sslobj, 'compression'):
                return self._sslobj.compression()
            else:
                return _sslcompat.compression(self._sslobj)

        def pending(self):
            return self._sslobj.pending()

        def do_handshake(self, block=False):
            self._sslobj.do_handshake()
            if getattr(self.context, 'check_hostname', False):
                if not self.server_hostname:
                    raise ValueError('check_hostname needs server_hostname argument')
                ssl.match_hostname(self.getpeercert(), self.server_hostname)

        def unwrap(self):
            self._sslobj.shutdown()

        def get_channel_binding(self, cb_type="tls-unique"):
            if cb_type not in CHANNEL_BINDING_TYPES:
                raise ValueError('Unsupported channel binding type')
            if cb_type != 'tls-unique':
                raise NotImplementedError('{0} channel binding not implemented'.format(cb_type))
            if hasattr(self._sslobj, 'tls_unique_cb'):
                return self._sslobj.tls_unique_cb()
            else:
                return _sslcompat.tls_unique_cb(self._sslobj)


def get_dummy_socket():
    # Return a dummy socket that can be wrapped by _ssl.
    sock = socket.socket()
    if six.PY2:
        sock = sock._sock
    sock.setblocking(False)
    return sock

_sock = None


def wrap_bio(ctx, incoming, outgoing, server_side=False, server_hostname=None):
    """Create a new SSL protocol instance from a context and a BIO pair."""
    if hasattr(ctx, 'wrap_bio'):
        # Python 3.5
        return ctx.wrap_bio(incoming, outgoing, server_side, server_hostname)
    # Allocate a single global dummy socket to wrap on Python < 3.5
    global _sock
    if _sock is None:
        _sock = get_dummy_socket()
    if hasattr(ctx, '_wrap_socket'):
        # Python 2.7.9+, Python 3.x
        sslobj = ctx._wrap_socket(_sock, server_side, server_hostname)
    else:
        # Python 2.7.x, x <= 8
        #
        # We need to do some magic to support anonymous DH authentication. Anon
        # DH doesn't need a certfile and a keyfile, but the Python 2.x
        # _ssl.sslwrap raises an exception if these are absent for server side
        # sockets. So the workaround is to create the socket as a client-side
        # socket and then flip it afterwards if needed.
        sslobj = _ssl.sslwrap(_sock, False, *ctx._ssl_args)
        if ctx._dh_params:
            _sslcompat.load_dh_params(sslobj, ctx._dh_params)
        if server_side:
            _sslcompat.set_accept_state(sslobj)
        _sslcompat.set_options(sslobj, ctx._options)
        if server_hostname:
            _sslcompat.set_tlsext_host_name(server_hostname)
    _sslcompat.set_bio(sslobj, incoming, outgoing)
    pyobj = SSLObject(sslobj, ctx, server_side, server_hostname)
    return pyobj


def get_reason(exc):
    """Return the reason code from an SSLError exception."""
    # On Python 3.x the reason is available via exc.reason.
    if hasattr(exc, 'reason'):
        return exc.reason
    # .. but on 2.x we have to parse the error string (fortunately it is there)
    message = exc.args[1]
    p0 = message.find('error:')
    if p0 == -1:
        return
    p0 += 6
    p1 = message.find(':', p0)
    assert p1 != -1
    code = int(message[p0:p1], 16) & 0xfff
    return errorcode.get(code)



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2017 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import socket
import unittest
import pyuv

from support import UnitTest
from gruvi.transports import TransportError, Transport, DatagramTransport


class ProtocolLogger(object):
    """Utility protocol class that implements both the stream and datagram
    protocols and that logs all callbacks."""

    def __init__(self):
        self.events = []
        self.transport = None

    def get_events(self, typ):
        return [ev for ev in self.events if ev[0] == typ]

    def connection_made(self, transport):
        self.events.append(('connection_made', transport))
        self.transport = transport

    def data_received(self, data):
        # collapse data_received for easier verification
        if self.events[-1][0] == 'data_received':
            self.events[-1] = ('data_received', self.events[-1][1] + data)
        else:
            self.events.append(('data_received', data))

    def eof_received(self):
        self.events.append(('eof_received',))

    def connection_lost(self, exc=None):
        self.events.append(('connection_lost', exc))

    def datagram_received(self, data, addr):
        self.events.append(('datagram_received', data, addr))

    def error_received(self, exc):
        self.events.append(('error_received', exc))

    def resume_writing(self):
        pass

    def pause_writing(self):
        pass


class EchoServer(ProtocolLogger):

    def data_received(self, data):
        super(EchoServer, self).data_received(data)
        self.transport.write(data)


class EventLoopTest(UnitTest):

    def setUp(self):
        super(EventLoopTest, self).setUp()
        self.loop = pyuv.Loop()
        self.errors = []

    def catch_errors(self, callback):
        # Run *callback*. If it raises an exception, store it and stop the
        # loop.
        def run_callback(*args):
            try:
                callback(*args)
            except Exception as e:
                self.errors.append(e)
                self.loop.stop()
        return run_callback

    def run_loop(self, timeout):
        # Run the loop for at most *timeout* seconds. Re-raise any exception
        # that was catch by a "catch_errors" callback.
        timer = pyuv.Timer(self.loop)
        def stop_loop(handle):
            self.loop.stop()
        timer.start(stop_loop, timeout, 0)
        self.loop.run()
        if self.errors:
            raise self.errors[0]


class TransportTest(object):

    def create_handle(self):
        raise NotImplementedError

    def bind_handle(self, handle):
        raise NotImplementedError

    def create_transport(self, handle, protocol, server_side):
        transport = Transport(handle)
        transport.start(protocol)
        return transport

    def test_echo(self):
        # Test a simple echo server. The client writes some data end then
        # writes an EOF (if the transport supports writing EOF). The server
        # echos and upon receipt of EOF will close the connection.
        @self.catch_errors
        def echo_server(handle, error):
            if error:
                raise TransportError.from_errno(error)
            client = self.create_handle()
            handle.accept(client)
            protocols[0] = EchoServer()
            transports[0] = self.create_transport(client, protocols[0], True)
        @self.catch_errors
        def echo_client(handle, error):
            if error:
                raise TransportError.from_errno(error)
            protocols[1] = ProtocolLogger()
            trans = transports[1] = self.create_transport(handle, protocols[1], False)
            trans.write(b'foo\n')
            trans.write(b'bar\n')
            trans.writelines([b'qux', b'quux'])
            if trans.can_write_eof():
                trans.write_eof()
        transports = [None, None]
        protocols = [None, None]
        server = self.create_handle()
        addr = self.bind_handle(server)
        server.listen(echo_server)
        client = self.create_handle()
        client.connect(addr, echo_client)
        self.run_loop(0.1)
        strans, ctrans = transports
        sproto, cproto = protocols
        self.assertIsInstance(strans, Transport)
        self.assertIsInstance(ctrans, Transport)
        self.assertIsInstance(sproto, EchoServer)
        self.assertIsInstance(cproto, ProtocolLogger)
        ctrans.close()
        self.run_loop(0.1)
        sevents = sproto.events
        self.assertIn(len(sevents), (3, 4))
        self.assertEqual(sevents[0], ('connection_made', strans))
        self.assertEqual(sevents[1], ('data_received', b'foo\nbar\nquxquux'))
        if strans.can_write_eof():
            self.assertEqual(sevents[2], ('eof_received',))
        self.assertEqual(sevents[-1][0], 'connection_lost')
        cevents = cproto.events
        self.assertIn(len(cevents), (3, 4))
        self.assertEqual(cevents[0], ('connection_made', ctrans))
        self.assertEqual(cevents[1], ('data_received', b'foo\nbar\nquxquux'))
        if ctrans.can_write_eof():
            self.assertEqual(cevents[2], ('eof_received',))
        self.assertEqual(cevents[-1][0], 'connection_lost')


class TestTcpTransport(TransportTest, EventLoopTest):

    def create_handle(self):
        return pyuv.TCP(self.loop)

    def bind_handle(self, handle):
        host = socket.gethostbyname('localhost')
        handle.bind((host, 0))
        return handle.getsockname()


class TestPipeTransport(TransportTest, EventLoopTest):

    def create_handle(self):
        return pyuv.Pipe(self.loop)

    def bind_handle(self, handle):
        addr = self.pipename('test-pipe')
        handle.bind(addr)
        return addr


class TestUdpTransport(EventLoopTest):

    def create_handle(self):
        return pyuv.UDP(self.loop)

    def bind_handle(self, handle):
        host = socket.gethostbyname('localhost')
        handle.bind((host, 0))
        return handle.getsockname()

    def create_transport(self, handle, protocol):
        transport = DatagramTransport(handle)
        transport.start(protocol)
        return transport

    def test_echo(self):
        server = self.create_handle()
        saddr = self.bind_handle(server)
        sproto = ProtocolLogger()
        strans = self.create_transport(server, sproto)
        client = self.create_handle()
        caddr = self.bind_handle(client)
        cproto = ProtocolLogger()
        ctrans = self.create_transport(client, cproto)
        # Try 5 times (since UDP is lossy)
        for i in range(5):
            ctrans.sendto(b'foo', saddr)
        for i in range(5):
            strans.sendto(b'bar', caddr)
        self.run_loop(0.1)
        sevents = sproto.get_events('datagram_received')
        self.assertGreater(len(sevents), 0)
        for event in sevents:
            self.assertEqual(event[1], b'foo')
            self.assertEqual(event[2], caddr)
        cevents = cproto.get_events('datagram_received')
        self.assertGreater(len(cevents), 0)
        for event in cevents:
            self.assertEqual(event[1], b'bar')
            self.assertEqual(event[2], saddr)


if __name__ == '__main__':
    unittest.main()



"""
This module implements DBus authentication mechanisms

@author: Tom Cocagne
"""

from __future__ import absolute_import, print_function

import os
import os.path
import time
import getpass
import hashlib
import binascii
import six

from   .error    import DBusAuthenticationFailed

class log:
    from gruvi.logging import get_logger
    logger = get_logger()
    @classmethod
    def msg(cls, m):
        cls.logger.warning(m)


def hexlify(s):
    # Like binascii.hexlify but returns a str
    if isinstance(s, six.text_type):
        s = s.encode('ascii')
    return binascii.hexlify(s).decode('ascii')


class ClientAuthenticator (object):
    """
    Implements the client-side portion of the DBus authentication protocol.

    @ivar preference: List of authentication mechanisms to try in the preferred order
    @type preference: List of C{string}
    """

    preference = ['EXTERNAL', 'DBUS_COOKIE_SHA1', 'ANONYMOUS']
    cookie_dir    = None # used for testing only
    
    def beginAuthentication(self, protocol):
        self.authenticated = False
        self.protocol      = protocol
        self.guid          = None

        self.authOrder = self.preference[:]
        self.authOrder.reverse()
        
        self.authTryNextMethod()
        

    def handleAuthMessage(self, line):
        if not ' ' in line:
            cmd = line
            args = ''
        else:
            cmd, args = line.split(' ',1)
        m = getattr(self, '_auth_' + cmd, None)
        if m:
            m( args )
        else:
            raise DBusAuthenticationFailed('Invalid DBus authentcation protocol message: ' + line)


    def authenticationSucceeded(self):
        return self.authenticated

    
    def getGUID(self):
        return self.guid
    
    #---------------------------------------------------

    def sendAuthMessage(self, msg):
        self.protocol.sendAuthMessage( msg )
        
            
    def authTryNextMethod(self):
        """
        Tries the next authentication method or raises a failure if all mechanisms
        have been tried.
        """
        if not self.authOrder:
            raise DBusAuthenticationFailed()
        
        self.authMech  = self.authOrder.pop()
            
        if self.authMech == 'DBUS_COOKIE_SHA1':
            self.sendAuthMessage('AUTH ' + self.authMech + ' ' +
                                 hexlify(getpass.getuser()))
        elif self.authMech == 'ANONYMOUS':
            self.sendAuthMessage('AUTH ' + self.authMech + ' ' +
                                 hexlify("txdbus"))
        else:
            self.sendAuthMessage('AUTH ' + self.authMech)

                    
    def _auth_REJECTED(self, line):
        self.authTryNextMethod()

        
    def _auth_OK(self, line):
        line = line.strip()

        if not line:
            raise DBusAuthenticationFailed('Missing guid in OK message')
        
        try:
            self.guid = line
        except:
            raise DBusAuthenticationFailed('Invalid guid in OK message')
        else:
            self.sendAuthMessage('BEGIN')
            self.authenticated = True
        

    def _auth_AGREE_UNIX_FD(self, line):
        log.msg('DBus Auth not implemented AGREE_UNIX_FD')
        
    
    def _auth_DATA(self, line):
        
        if self.authMech == 'EXTERNAL':
            self.sendAuthMessage('DATA')
            
        elif self.authMech == 'DBUS_COOKIE_SHA1':
            try:
                data = binascii.unhexlify( line.strip() ).decode('ascii')
                
                cookie_context, cookie_id, server_challenge = data.split()

                server_cookie = self._authGetDBusCookie(cookie_context, cookie_id)

                client_challenge = hexlify(hashlib.sha1(os.urandom(8)).digest())

                response = '%s:%s:%s' % (server_challenge,
                                         client_challenge,
                                         server_cookie)

                response = hexlify(hashlib.sha1(response.encode('ascii')).digest())

                reply = client_challenge + ' ' + response
                
                self.sendAuthMessage( 'DATA ' + hexlify(reply))
            except Exception as e:
                log.msg('DBUS Cookie authentication failed: ' + str(e))
                self.sendAuthMessage('ERROR ' + str(e))

    def _auth_ERROR(self, line):
        log.msg('Authentication mechanism failed: ' + line)
        self.authTryNextMethod()

    #--------------------------------------------------
    
    def _authGetDBusCookie(self, cookie_context, cookie_id):
        """
        Reads the requested cookie_id from the cookie_context file
        """
        # XXX   Ensure we obtain the correct directory for the
        #       authenticating user and that that user actually
        #       owns the keyrings directory

        if self.cookie_dir is None:
            cookie_dir = os.path.expanduser('~/.dbus-keyrings')
        else:
            cookie_dir = self.cookie_dir

        dstat = os.stat(cookie_dir)

        if dstat.st_mode & 0x36:  # 066
            raise Exception('User keyrings directory is writeable by other users. Aborting authentication')

        import pwd
        if dstat.st_uid != pwd.getpwuid(os.geteuid()).pw_uid:
            raise Exception('Keyrings directory is not owned by the current user. Aborting authentication!')
        
        f = open(os.path.join(cookie_dir, cookie_context), 'r')

        try:
            for line in f:
                try:
                    k_id, k_time, k_cookie_hex = line.split()
                    if k_id == cookie_id:
                        return k_cookie_hex
                except:
                    pass
        finally:
            f.close()


class BusCookieAuthenticator (object):
    """
    Implements the Bus-side portion of the DBUS_COOKIE_SHA1 authentication
    mechanism
    """
    
    cookieContext = 'org_twisteddbus_ctx' + str(os.getpid())
    keyring_dir = None

    
    def __init__(self):
        self.step_num = 0
        self.username = None
        self.cookieId = None

        
    def cancel(self):
        if self.cookieId:
            self._delete_cookie()


    def getMechanismName(self):
        return 'DBUS_COOKIE_SHA1'


    def init(self, protocol):
        pass

    
    def getUserName(self):
        return self.username
    
    
    def step(self, arg):
        s = self.step_num
        self.step_num += 1

        if arg is None:
            return ('REJECTED', None)

        try:
            if s == 0:
                return self._step_one( arg )
            elif s == 1:
                return self._step_two( arg )
            else:
                raise Exception()
        except Exception as e:
            return ('REJECTED', None)

        
    def _step_one(self, username, keyring_dir=None):
        try:
            uid = int(username)
            try:
                import pwd
                username = pwd.getpwuid(uid).pw_name
            except:
                return ('REJECTED', None)
        except ValueError:
            pass
        
        self.username = username
        
        try:
            import pwd
            p = pwd.getpwnam(username)
            self.uid     = p.pw_uid
            self.gid     = p.pw_gid
            self.homedir = p.pw_dir
        except (KeyError, ImportError):
            return ('REJECTED', None) # username not found

        if keyring_dir is None:
            keyring_dir = self.keyring_dir

        if keyring_dir is None:
            dk = os.path.join(self.homedir, '.dbus-keyrings')
        else:
            dk = keyring_dir # for testing only
        
        self.cookie_file = os.path.join(dk, self.cookieContext)
        
        try:
            s = os.lstat(dk)
            if not os.path.isdir(dk) or s.st_mode & 0x36:  # 066
                # Invalid keyrings directory. Something fishy is going on
                return ('REJECTED', None)
        except OSError:
            old_un = os.umask(0x3f)  # 077
            os.mkdir(dk)
            os.umask(old_un)
            if os.geteuid() == 0:
                os.chown(dk, self.uid, self.gid)

        self._create_cookie()

        self.challenge_str = hexlify(hashlib.sha1(os.urandom(8)).digest())
        
        msg = ' '.join( [self.cookieContext,
                         str(self.cookieId),
                         self.challenge_str] )
        
        return ('CONTINUE', msg)

    
    def _step_two(self, response):
        self._delete_cookie()
        hash_str = None
        shash    = 1
        try:
            client_challenge, hash_str = response.split()

            tohash = self.challenge_str + ':' + client_challenge + ':' + self.cookie

            shash = hexlify( hashlib.sha1(tohash.encode('ascii')).digest() )
        except:
            pass

        if shash == hash_str:
            return ('OK', None)
        else:
            return ('REJECTED', None)

        

    def _get_lock(self):
        #
        # Twisted is single-threaded, our context name includes the
        # process id, and we wont be using any deferreds here... so
        # the lock file really isn't needed. We'll go ahead and
        # include a very simple version of it, however, "just to say
        # we did"
        #
        self.lock_file = self.cookie_file + '.lock'
        try:
            lockfd = os.open(self.lock_file, os.O_CREAT | os.O_EXCL
                                             | os.O_WRONLY, 0x180)  # 0600
        except:
            time.sleep(0.01)
            try:
                lockfd = os.open(self.lock_file, os.O_CREAT | os.O_EXCL
                                                 | os.O_WRONLY, 0x180)  # 0600
            except:
                os.unlink(self.lock_file)
                lockfd = os.open(self.lock_file, os.O_CREAT | os.O_EXCL
                                                 | os.O_WRONLY, 0x180)  # 0600

        return lockfd


    def _get_cookies(self,timefunc=time.time):
        cookies = list()
        f = None
        try:
            f = open(self.cookie_file, 'r')
            for line in f:
                k_id, k_time, k_cookie_hex = line.split()

                if abs( timefunc() - int(k_time) ) < 30:
                    cookies.append( line.split() )
        except:
            pass
        finally:
            if f:
                f.close()

        return cookies
    
    
    def _create_cookie(self, timefunc=time.time):
        
        lockfd = self._get_lock()

        cookies = self._get_cookies(timefunc)

        cookie_id = 1
        for tpl in cookies:
            if int(tpl[0]) >= cookie_id:
                cookie_id = int(tpl[0]) + 1

        cookie = hexlify(os.urandom(24))
        
        cookies.append( (str(cookie_id), str(int(timefunc())), cookie) )

        for c in cookies:
            os.write(lockfd, ' '.join(c).encode('ascii') + b'\n')

        os.close(lockfd)
        if os.geteuid() == 0:
            os.chown(self.lock_file, self.uid, self.gid)

        os.rename(self.lock_file, self.cookie_file)

        self.cookieId = cookie_id
        self.cookie   = cookie

    
    def _delete_cookie(self):
        lockfd = self._get_lock()

        cookies = self._get_cookies()

        for i, tpl in enumerate(cookies):
            if int(tpl[0]) == self.cookieId:
                del cookies[i]
                break
            
        if not cookies:
            os.unlink(self.cookie_file)
            os.close(lockfd)
            os.unlink(self.lock_file)
        else:
            for c in cookies:
                os.write(lockfd, ' '.join(c).encode('ascii') + b'\n')

            os.close(lockfd)
            if os.geteuid() == 0:
                os.chown(self.lock_file, self.uid, self.gid)
            
            os.rename(self.lock_file, self.cookie_file) 
        

class BusExternalAuthenticator (object):
    """
    Implements the Bus-side portion of the EXTERNAL authentication
    mechanism
    """

    def __init__(self):
        self.ok = False
        self.creds = None
        
    def getMechanismName(self):
        return 'EXTERNAL'

    def init(self, protocol):
        self.creds = protocol._unix_creds
    
    def step(self, arg):
        if not self.creds:
            return ('REJECT', 'Unix credentials not available')
        if not self.ok:
            self.ok = True
            return ('CONTINUE', '')
        else:
            return ('OK', None)

    def getUserName(self):
        import pwd
        return pwd.getpwuid(self.creds[1]).pw_name
    

    def cancel(self):
        pass


class BusAnonymousAuthenticator (object):
    """
    Implements the Bus-side portion of the ANONYMOUS authentication
    mechanism
    """
    
    def getMechanismName(self):
        return 'ANONYMOUS'

    def init(self, protocol):
        pass
    
    def step(self, arg):
        return ('OK', None)

    def getUserName(self):
        return 'anonymous'

    def cancel(self):
        pass
    

class BusAuthenticator (object):
    """
    Implements the Bus-side portion of the DBus authentication protocol.

    @ivar authenticators: A dictionary of mechanism names to mechanism
                          implementation classes
    @type authenticators: C{dict}
    """

    MAX_REJECTS_ALLOWED = 5

    authenticators = { 'EXTERNAL'         : BusExternalAuthenticator,
                       'DBUS_COOKIE_SHA1' : BusCookieAuthenticator,
                       'ANONYMOUS'        : BusAnonymousAuthenticator }
    
    def __init__(self, server_guid):
        self.server_guid   = server_guid
        self.authenticated = False
        self.mechanisms    = dict()
        self.protocol      = None
        self.guid          = None

        self.reject_count  = 0
        self.state         = None
        self.current_mech  = None

        for n, m in six.iteritems(self.authenticators):
            self.mechanisms[ n ] = m

        mechNames = self.authenticators.keys()
        
        self.reject_msg = 'REJECTED ' + ' '.join(mechNames)


        
    def beginAuthentication(self, protocol):
        self.protocol      = protocol
        self.state         = 'WaitingForAuth'
        

    def handleAuthMessage(self, line):
        #print 'RCV: ', line.rstrip()
        if not ' ' in line:
            cmd = line
            args = ''
        else:
            cmd, args = line.split(' ',1)
        m = getattr(self, '_auth_' + cmd, None)
        if m:
            m( args )
        else:
            self.sendError('"Unknown command"')


    def authenticationSucceeded(self):
        return self.authenticated

    
    def getGUID(self):
        return self.server_guid
    
    #---------------------------------------------------

    def reject(self):
        if self.current_mech:
            self.current_mech.cancel()
            self.current_mech = None
            
        self.reject_count += 1
        if self.reject_count > self.MAX_REJECTS_ALLOWED:
            raise DBusAuthenticationFailed('Client exceeded maximum failed authentication attempts')
        
        self.sendAuthMessage(self.reject_msg)
        self.state = 'WaitingForAuth'
        
    
    def sendAuthMessage(self, msg):
        #print 'SND: ', msg.rstrip()
        self.protocol.sendAuthMessage( msg )

        
    def sendError(self, msg = None):
        if msg:
            self.sendAuthMessage('ERROR ' + msg)
        else:
            self.sendAuthMessage('ERROR')


    def stepAuth(self, response):
        if self.current_mech is None:
            self.reject()
            return

        if response:
            response = binascii.unhexlify( response.strip() ).decode('ascii')

            
        status, challenge = self.current_mech.step(response)


        if status == 'OK':
            self.sendAuthMessage('OK ' + self.server_guid)
            self.state = 'WaitingForBegin'

        elif status == 'CONTINUE':
            self.sendAuthMessage('DATA ' + hexlify(challenge))
            self.state = 'WaitingForData'
            
        else:
            #print 'REJECT: ', status
            self.reject()


    def _auth_AUTH(self, line):
        if self.state == 'WaitingForAuth':
            tpl = line.split()

            if len(tpl) == 0:
                self.reject()
            else:
                mech             = tpl[0]
                initial_response = None
                if len(tpl) > 1:
                    initial_response = tpl[1]
                if mech in self.mechanisms:
                    self.current_mech = self.mechanisms[mech]()
                    self.current_mech.init(self.protocol)
                    self.stepAuth(initial_response)                    
                else:
                    self.reject()

        else:
            self.sendError()
                    
                    
    def _auth_BEGIN(self, line):
        if self.state == 'WaitingForBegin':
            self.authenticated = True
            self.guid          = self.current_mech.getUserName()
        else:
            raise DBusAuthenticationFailed('Protocol violation')
        
        
    def _auth_ERROR(self, line):
        if self.state in ('WaitingForAuth', 'WaitingForData',
                          'WaitingForBegin'):
            self.reject()


    def _auth_DATA(self, line):
        if self.state == 'WaitingForData':
            self.stepAuth(line)
        else:
            self.sendError()


    def _auth_CANCEL(self, line):
        if self.state in ('WaitingForData', 'WaitingForBegin'):
            self.reject()
        else:
            self.sendError()


    def _auth_NEGOTIATE_UNIX_FD(self, line):
        # Only valid in the 'WaitingForBegin' state
        self.sendError()

            



# Gruvi example program: echo server, using StreamServer

import gruvi

def echo_handler(stream, transport, protocol):
    peer = transport.get_extra_info('peername')
    print('New connection from {0}'.format(gruvi.saddr(peer)))
    while True:
        buf = stream.read1()
        if not buf:
            break
        stream.write(buf)
    print('Connection lost')

server = gruvi.StreamServer(echo_handler)
server.listen(('localhost', 0))
for addr in server.addresses:
    print('Listen on {0}'.format(gruvi.saddr(addr)))

server.run()



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2013 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import weakref
import fibers

__all__ = ['local']


class local(object):
    """Fiber-local data.

    To manage fiber-local data, instantiate this class and store attributes on
    it::

        mydata = local()
        mydata.x = 10

    Attributes have a value or are unset independently for each fiber.
    """

    def __init__(self):
        self.__dict__['_keys'] = weakref.WeakKeyDictionary()

    def __getattr__(self, key):
        current = fibers.current()
        try:
            return self._keys[current][key]
        except KeyError:
            raise AttributeError(key)

    def __setattr__(self, key, value):
        current = fibers.current()
        self._keys.setdefault(current, {})[key] = value

    def __delattr__(self, key):
        current = fibers.current()
        try:
            del self._keys[current][key]
        except KeyError:
            raise AttributeError(key)



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2017 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import os.path
from cffi import FFI

parent, _ = os.path.split(os.path.abspath(__file__))
topdir, _ = os.path.split(parent)


ffi = FFI()

ffi.set_source('http_ffi', """
    #include <stdlib.h>
    #include "src/http_parser.h"
    #include "src/http_parser.c"

    unsigned char http_message_type(http_parser *p) { return p->type; }
    unsigned int http_status_code(http_parser *p) { return p->status_code; }
    unsigned int http_method(http_parser *p) { return p->method; }
    unsigned char http_errno(http_parser *p) { return p->http_errno; }
    unsigned char http_is_upgrade(http_parser *p) { return p->upgrade; }
""", include_dirs=[topdir])

ffi.cdef("""
    typedef struct http_parser http_parser;
    typedef struct http_parser_settings http_parser_settings;

    typedef int (*http_data_cb) (http_parser*, const char *at, size_t length);
    typedef int (*http_cb) (http_parser*);

    enum http_parser_type { HTTP_REQUEST, HTTP_RESPONSE, HTTP_BOTH, ... };

    struct http_parser {
      unsigned short http_major;
      unsigned short http_minor;
      void *data;
      ...;
    };

    struct http_parser_settings {
      http_cb      on_message_begin;
      http_data_cb on_url;
      http_data_cb on_status;
      http_data_cb on_header_field;
      http_data_cb on_header_value;
      http_cb      on_headers_complete;
      http_data_cb on_body;
      http_cb      on_message_complete;
      ...;
    };

    enum http_parser_url_fields { UF_SCHEMA, UF_HOST, UF_PORT, UF_PATH,
            UF_QUERY, UF_FRAGMENT, UF_USERINFO, UF_MAX };

    struct http_parser_url {
        uint16_t field_set;
        uint16_t port;

        struct {
            uint16_t off;
            uint16_t len;
        } field_data[UF_MAX];
        ...;
    };

    void http_parser_init(http_parser *parser, enum http_parser_type type);
    size_t http_parser_execute(http_parser *parser,
                               const http_parser_settings *settings,
                               const char *data,
                               size_t len);

    int http_should_keep_alive(const http_parser *parser);
    const char *http_method_str(enum http_method m);
    const char *http_errno_name(enum http_errno err);

    void http_parser_url_init(struct http_parser_url *u);
    int http_parser_parse_url(const char *buf, size_t buflen,
                              int is_connect, struct http_parser_url *u);

    /* Extra functions to extract bitfields not supported by cffi */
    unsigned char http_message_type(http_parser *parser);
    unsigned int http_status_code(http_parser *parser);
    unsigned int http_method(http_parser *parser);
    unsigned char http_errno(http_parser *parser);
    unsigned char http_is_upgrade(http_parser *parser);

""")


if __name__ == '__main__':
    ffi.compile()



# Gruvi example program: an echo server, using a Protocol

import gruvi

class EchoProtocol(gruvi.Protocol):

    def connection_made(self, transport):
        self._transport = transport
        peer = transport.get_extra_info('peername')
        print('New connection from {0}'.format(gruvi.saddr(peer)))

    def data_received(self, data):
        self._transport.write(data)

    def eof_received(self):
        print('Connection lost')

server = gruvi.create_server(EchoProtocol, ('localhost', 0))
for addr in server.addresses:
    print('Listen on {0}'.format(gruvi.saddr(addr)))

server.run()



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2017 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import os
import sys
import signal
import unittest

from support import TestCase

import gruvi
from gruvi import Process, PIPE, DEVNULL, StreamClient, HttpClient


class TestExamples(TestCase):
    # Tests for the examples in examples/*

    @classmethod
    def setUpClass(cls):
        super(TestExamples, cls).setUpClass()
        exampledir = os.path.join(cls.topdir, 'examples')
        os.chdir(exampledir)

    def test_curl(self):
        proc = Process(encoding='utf-8')
        proc.spawn([sys.executable, 'curl.py', 'http://www.python.org'], stdout=PIPE)
        stdout, stderr = proc.communicate(timeout=30)
        self.assertEqual(proc.returncode, 0)
        self.assertEqual(stdout, '')
        proc.close()

    def test_curl_ssl(self):
        proc = Process(encoding='utf-8')
        proc.spawn([sys.executable, 'curl.py', 'https://www.python.org'], stdout=PIPE)
        stdout, stderr = proc.communicate(timeout=30)
        self.assertEqual(proc.returncode, 0)
        self.assertTrue('<html' in stdout)
        proc.close()

    def test_echoserver1(self):
        proc = Process(encoding='ascii')
        # -u: unbuffered stdio
        proc.spawn([sys.executable, '-u', 'echoserver1.py'], stdout=PIPE, stderr=DEVNULL)
        line = proc.stdout.readline()
        self.assertTrue(line.startswith('Listen on '))
        addr = gruvi.paddr(line[10:])
        client = StreamClient()
        client.connect(addr)
        client.write(b'foo')
        self.assertEqual(client.read(3), b'foo')
        client.write(b'foo bar baz\n')
        self.assertEqual(client.readline(), b'foo bar baz\n')
        client.close()
        proc.send_signal(signal.SIGINT)
        proc.wait(timeout=2)
        self.assertEqual(proc.returncode, 0)
        proc.close()

    def test_echoserver2(self):
        proc = Process(encoding='ascii')
        proc.spawn([sys.executable, '-u', 'echoserver2.py'], stdout=PIPE, stderr=DEVNULL)
        line = proc.stdout.readline()
        self.assertTrue(line.startswith('Listen on '))
        addr = gruvi.paddr(line[10:])
        client = StreamClient()
        client.connect(addr)
        client.write(b'foo')
        self.assertEqual(client.read(3), b'foo')
        client.write(b'foo bar baz\n')
        self.assertEqual(client.readline(), b'foo bar baz\n')
        client.close()
        proc.send_signal(signal.SIGINT)
        proc.wait(timeout=1)
        self.assertEqual(proc.returncode, 0)
        proc.close()

    def test_fortune(self):
        proc = Process(encoding='ascii')
        proc.spawn([sys.executable, '-u', 'fortune.py'], stdout=PIPE)
        line = proc.stdout.readline()
        self.assertTrue(line.startswith('Listen on '))
        addr = gruvi.paddr(line[10:])
        client = HttpClient()
        client.connect(addr)
        client.request('GET', '/')
        resp = client.getresponse()
        fortune = resp.body.read().decode('ascii')
        client.close()
        self.assertTrue('Albert Einstein' in fortune)
        proc.send_signal(signal.SIGINT)
        proc.wait(timeout=1)
        self.assertEqual(proc.returncode, 0)
        proc.close()

    def test_netcat(self):
        proc = Process()
        proc.spawn([sys.executable, '-u', 'netcat.py', 'python.org', '80'], stdin=PIPE, stdout=PIPE)
        proc.stdin.write(b'GET / HTTP/1.1\r\nHost: www.python.org\r\nConnection: close\r\n\r\n')
        result = proc.stdout.read()
        self.assertTrue(b'HTTP/1.1 301' in result)
        proc.wait(timeout=1)
        self.assertEqual(proc.returncode, 0)
        proc.close()

    def test_netcat_ssl(self):
        proc = Process()
        proc.spawn([sys.executable, '-u', 'netcat.py', '--ssl', 'python.org', '443'],
                   stdin=PIPE, stdout=PIPE)
        proc.stdin.write(b'GET / HTTP/1.1\r\nHost: www.python.org\r\nConnection: close\r\n\r\n')
        result = proc.stdout.read()
        self.assertTrue(b'<html' in result)
        proc.wait(timeout=1)
        self.assertEqual(proc.returncode, 0)
        proc.close()

    def test_jsonrpc(self):
        proc = Process(encoding='utf-8')
        proc.spawn([sys.executable, 'jsonrpc.py'], stdout=PIPE)
        result = proc.stdout.read()
        self.assertEqual(result, 'result = pong\n')
        proc.wait(timeout=1)
        self.assertEqual(proc.returncode, 0)
        proc.close()

    def test_http_cs(self):
        proc = Process(encoding='utf-8')
        proc.spawn([sys.executable, 'httpcs.py'], stdout=PIPE)
        result = proc.stdout.read()
        self.assertEqual(result, 'result = pong\n')
        proc.wait(timeout=1)
        self.assertEqual(proc.returncode, 0)
        proc.close()


if __name__ == '__main__':
    unittest.main()



#
# This file is part of gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2017 the gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import signal
import collections
import threading
import traceback
import functools

import pyuv
import fibers

from . import logging, compat
from .errors import Timeout
from .callbacks import add_callback, run_callbacks
from .poll import Poller

__all__ = ['switchpoint', 'assert_no_switchpoints', 'switch_back', 'get_hub',
           'Hub', 'sleep']


def switchpoint(func):
    """Mark *func* as a switchpoint.

    In Gruvi, all methods and functions that call :meth:`Hub.switch` directly,
    and all public APIs that can cause an indirect switch, are marked as a
    switchpoint. It is recommended that you mark your own methods and functions
    in the same way. Example::

      @switchpoint
      def myfunc():
          # may call Hub.switch() here
    """
    @functools.wraps(func)
    def switchpoint(*args, **kwargs):
        hub = get_hub()
        if fibers.current() is hub:
            raise RuntimeError('cannot call switchpoint from the Hub')
        if hub._noswitch_depth:
            raise AssertionError('switchpoint called from no-switch section')
        return func(*args, **kwargs)
    switchpoint.__switchpoint__ = True
    # The py27 version of functools.wraps() doesn't store the wrapped function
    # under __wrapped__. We need it as some of our methods call it in certain
    # cases if we are sure we're not going to block, e.g. Queue.put_nowait().
    if not hasattr(switchpoint, '__wrapped__'):
        switchpoint.__wrapped__ = func
    return switchpoint


class assert_no_switchpoints(object):
    """Context manager that defines a block in which no switches may happen,
    and in which no switchpoints may be called.

    Use it as follows::

      with assert_no_switchpoints():
          do_something()
          do_something_else()

    If the context manager detects a switch or a call into a switchpoint it
    raises an :exc:`AssertionError`.
    """

    __slots__ = ('_hub',)

    def __init__(self, hub=None):
        self._hub = hub or get_hub()

    def __enter__(self):
        self._hub._noswitch_depth += 1

    def __exit__(self, *exc_info):
        assert self._hub._noswitch_depth > 0
        self._hub._noswitch_depth -= 1
        self._hub = None


class switch_back(object):
    """A context manager to facilitate switching back to the current fiber.

    Instances of this class are callable, and are intended to be used as the
    callback argument for an asynchronous operation. When called, the
    switchback object causes :meth:`Hub.switch` to return in the *origin* fiber
    (the fiber that created the switchback object). The return value in the
    origin fiber will be an ``(args, kwargs)`` tuple containing positional and
    keyword arguments passed to the callback.

    When the context manager exits it will be deactivated. If it is called
    after that then no switch will happen. Also the cleanup callbacks are run
    when the context manager exits.

    In the example below, a switchback object is used to wait for at most 10
    seconds for a SIGHUP signal::

      hub = get_hub()
      with switch_back(timeout=10) as switcher:
          sigh = pyuv.Signal(hub.loop)
          sigh.start(switcher, signal.SIGHUP)
          switcher.add_cleanup(sigh.close)
          hub.switch()
    """

    __slots__ = ('_timeout', '_hub', '_fiber', '_timer', '_callbacks', '_lock')

    def __init__(self, timeout=None, hub=None, lock=None):
        """
        The *timeout* argument can be used to force a timeout after this many
        seconds. It can be an int or a float. If a timeout happens,
        :meth:`Hub.switch` will raise a :class:`Timeout` exception in the
        origin fiber. The default is None, meaning there is no timeout.

        The *hub* argument can be used to specify an alternate hub to use.
        This argument is used by the unit tests and should normally not be
        needed.
        """
        self._timeout = timeout
        self._hub = hub or get_hub()
        self._fiber = fibers.current()
        self._callbacks = None
        self._lock = lock

    @property
    def fiber(self):
        """The origin fiber."""
        return self._fiber

    @property
    def timeout(self):
        """The :class:`~gruvi.Timeout` exception if a timeout has occurred.
        Otherwise the *timeout* parameter provided to the constructor."""
        return self._timeout

    @property
    def active(self):
        """Whether the switchback object is currently active."""
        return bool(self._hub and self._fiber.is_alive())

    def switch(self, value=None):
        """Switch back to the origin fiber. The fiber is switch in next time
        the event loop runs."""
        if self._hub is None or not self._fiber.is_alive():
            return
        self._hub.run_callback(self._fiber.switch, value)
        self._hub = self._fiber = None  # switch back at most once!

    def throw(self, typ, val=None, tb=None):
        """Throw an exception into the origin fiber. The exception is thrown
        the next time the event loop runs."""
        # The might seem redundant with self._fiber.cancel(exc), but it isn't
        # as self._fiber might be a "raw" fibers.Fiber() that doesn't have a
        # cancel() method.
        if self._hub is None or not self._fiber.is_alive():
            return
        self._hub.run_callback(self._fiber.throw, typ, val, tb)
        self._hub = self._fiber = None  # switch back at most once!

    def add_cleanup(self, callback, *args):
        """Add a cleanup action. The callback is run with the provided
        positional arguments when the context manager exists."""
        add_callback(self, callback, args)

    def __enter__(self):
        if self._timeout is not None:
            # There are valid scenarios for a Gruvi application where the loop
            # will not run for a long time. For example, a single fiber program
            # that only calls out to the loop to perform a blocking action.
            # Since a timer captures loop->time at when it is created, we need
            # to make sure the loop's time is up to date. That's why we call
            # update_time().
            self._hub.loop.update_time()
            self._timer = pyuv.Timer(self._hub.loop)
            self._timer.start(self, self._timeout, 0)
        return self

    def __exit__(self, *exc_info):
        if self._timeout is not None:
            self._timer.close()
            self._timer = None
        run_callbacks(self)

    def __call__(self, *args, **kwargs):
        # This method is thread safe if a lock was passed into the constructor
        # (and is the only method in this class for which this is the case).
        # This functionality is provided because this class is used by thread
        # safe objects where it can happen that the timeout and the regular
        # invocation of this method can happen in different threads.
        if self._lock:
            self._lock.acquire()
        try:
            if self._timeout is not None and args == (self._timer,):
                self._timeout = Timeout('timeout in switch_back() block')
                self.throw(Timeout, self._timeout)
            else:
                self.switch((args, kwargs))
        finally:
            if self._lock:
                self._lock.release()


_local = threading.local()

def get_hub():
    """Return the instance of the hub."""
    try:
        hub = _local.hub
    except AttributeError:
        hub = _local.hub = Hub()
    return hub


class Hub(fibers.Fiber):
    """The central fiber scheduler and event loop manager."""

    # The hub is created automatically the first time it is needed,so it is not
    # necessary to instantiate this class yourself.
    #
    # By default there is one hub per thread. To access the per thread instance,
    # use get_hub()
    #
    # The hub is used by fibers to pause themselves until a wake-up condition
    # becomes true. See the documentation for switch_back for details

    def __init__(self):
        if self.parent is not None:
            raise RuntimeError('Hub must be created in the root fiber')
        super(Hub, self).__init__(target=self.run)
        self.name = 'Hub'
        self.context = ''
        self._loop = pyuv.Loop()
        self._loop.excepthook = self._on_uncaught_exception
        self._data = {}
        self._noswitch_depth = 0
        # Use a deque() instead of our dllist from gruvi.callbacks because we
        # need a thread-safe append, and we don't need to remove things from
        # the middle.
        self._callbacks = collections.deque()
        # Thread IDs may be recycled when a thread exits. But as long as the
        # hub is alive, it won't be recycled so in that case we can use just
        # the ID as a check whether we are in the same thread or not.
        self._thread = compat.get_thread_ident()
        self._async_handle = pyuv.Async(self._loop, lambda h: self._loop.stop())
        self._sigint_handle = pyuv.Signal(self._loop)
        self._sigint_handle.start(self._on_sigint, signal.SIGINT)
        # Mark our own handles as "system handles". This allows the test suite
        # to check that no active handles except these escape from tests.
        self._async_handle._system_handle = True
        self._sigint_handle._system_handle = True
        self._log = logging.get_logger()
        self._log.debug('new Hub for {.name}', threading.current_thread())
        self._closing = False
        self._poll = Poller(self)

    @property
    def loop(self):
        """The event loop used by this hub instance. This is an instance of
        :class:`pyuv.Loop`."""
        return self._loop

    @property
    def data(self):
        """A per-hub dictionary that can be used by applications to store data.

        Keys starting with ``'gruvi:'`` are reserved for internal use."""
        return self._data

    @property
    def poll(self):
        """A centrally managed poller that can be used install callbacks for
        file descriptor readiness events."""
        return self._poll

    def _on_sigint(self, h, signo):
        # SIGINT handler. Terminate the hub and switch back to the root.
        self._log.debug('SIGINT received, stopping loop')
        self.close()

    def _on_uncaught_exception(self, *exc_info):
        # Installed as the handler for uncaught exceptions in pyuv callbacks.
        # The exception is cleared by pyuv when this method is called. So we
        # have to format *exc_info ourselves, we cannot use _log.exception().
        self._log.error('uncaught exception in pyuv callback')
        trace = '\n'.join(traceback.format_exception(*exc_info))
        self._log.error('Traceback (most recent call last):\n{}', trace)

    def _interrupt_loop(self):
        # Interrupt the event loop. If this is a cross-thread interrupt we need
        # to do this via the async handle.
        if compat.get_thread_ident() == self._thread:
            self._loop.stop()
        else:
            self._async_handle.send()

    def close(self):
        """Close the hub.

        This sets a flag that will cause the event loop to exit when it next
        runs. The hub fiber will then exit and control is transferred back to
        the root fiber.
        """
        if self._loop is None:
            return
        self._closing = True
        self._poll.close()
        self._interrupt_loop()

    def run(self):
        # Target of Hub.switch().
        if self.current() is not self:
            raise RuntimeError('run() may only be called from the Hub')
        self._log.debug('starting hub fiber')
        # This is where the loop runs. There are two loops:
        #
        # * An outer loop, in Python. This is where callbacks scheduled via
        #   run_callback() run. We call these "Python" callbacks, and they are
        #   alowed to switch between fibers.
        # * An inner loop, in libuv. This is where we block for input. Fiber
        #   switches are not allowed here, primarily to enforce that protocol
        #   callbacks stay fully event based and do not switch.
        #
        # The libuv loop is the main loop, and is interrupted only when Python
        # callbacks need to run (via _interrupt_loop()). After those are run,
        # we enter the libuv loop again.
        while True:
            self._run_callbacks()
            if self._closing:
                break
            # If the Python callbacks run above scheduled further callbacks, do
            # not wait for new events in the libuv loop. This would cause the
            # Pyton callbacks to potentially be delayed indefinitely.
            mode = pyuv.UV_RUN_NOWAIT if len(self._callbacks) else pyuv.UV_RUN_DEFAULT
            with assert_no_switchpoints(self):
                self._loop.run(mode)
        # Hub is going to exit at this point. Clean everyting up.
        for handle in self._loop.handles:
            if not handle.closed:
                handle.close()
        # Run the loop until all asynchronous closes are handled.
        # For some reason it appears this needs to be run twice.
        while self._loop.run():
            self._log.debug('run loop another time to close handles')
        if getattr(_local, 'hub', None) is self:
            del _local.hub
        self._loop = None
        self._callbacks.clear()
        self._async = None
        self._sigint = None
        self._log.debug('hub fiber terminated')

    def switch(self):
        """Switch to the hub.

        This method pauses the current fiber and runs the event loop. The
        caller should ensure that it has set up appropriate callbacks so that
        it will get scheduled again, preferably using :class:`switch_back`. In
        this case then return value of this method will be an ``(args,
        kwargs)`` tuple containing the arguments passed to the switch back
        instance.

        If this method is called from the root fiber then there are two
        additional cases. If the hub exited due to a call to :meth:`close`,
        then this method returns None. And if the hub exited due to a
        exception, that exception is re-raised here.
        """
        if self._loop is None or not self.is_alive():
            raise RuntimeError('hub is closed/dead')
        elif self.current() is self:
            raise RuntimeError('cannot switch to myself')
        elif compat.get_thread_ident() != self._thread:
            raise RuntimeError('cannot switch from a different thread')
        value = super(Hub, self).switch()
        if isinstance(value, Exception):
            raise value
        return value

    def _run_callbacks(self):
        """Run registered callbacks."""
        for i in range(len(self._callbacks)):
            callback, args = self._callbacks.popleft()
            try:
                callback(*args)
            except Exception:
                self._log.exception('Ignoring exception in callback:')

    def run_callback(self, callback, *args):
        """Queue a callback.

        The *callback* will be called with positional arguments *args* in the
        next iteration of the event loop. If you add multiple callbacks, they
        will be called in the order that you added them. The callback will run
        in the Hub's fiber.

        This method is thread-safe: it is allowed to queue a callback from a
        different thread than the one running the Hub.
        """
        if self._loop is None:
            raise RuntimeError('hub is closed')
        elif not callable(callback):
            raise TypeError('"callback": expecting a callable')
        self._callbacks.append((callback, args))  # thread-safe
        self._interrupt_loop()


@switchpoint
def sleep(secs):
    """Sleep for *secs* seconds. The *secs* argument can be an int or a float."""
    hub = get_hub()
    try:
        with switch_back(secs):
            hub.switch()
    except Timeout:
        pass



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import os
import sys
import time
import signal
import locale
import unittest

import pyuv
import gruvi
from gruvi.hub import get_hub
from gruvi.errors import Timeout
from gruvi.process import Process, PIPE, DEVNULL
from gruvi.stream import StreamClient

from support import UnitTest


def python_script(args):
    """On Windows, modify *args* so that a script in test/bin is executed
    directly via Python rather than via the .exe wrapper."""
    # This is needed to make inheriting of handles other than stdio work. My
    # assumption is that the .exe wrapper is not passing on these handles (or
    # that they are somehow not inheritable).
    if not sys.platform.startswith('win'):
        return args
    if isinstance(args, str):
        args = [args]
    args[0] = 'bin\\{0}'.format(args[0])
    return [sys.executable] + args


class TestProcess(UnitTest):

    def test_spawn(self):
        # Ensure that spawning a child works.
        proc = Process()
        proc.spawn('true')
        self.assertIsInstance(proc.pid, int)
        proc.wait()
        self.assertEqual(proc.returncode, 0)
        proc.close()

    def test_respawn(self):
        # Ensure that calling spawn() again after the child has exited works.
        proc = Process()
        proc.spawn('true')
        proc.wait()
        self.assertEqual(proc.returncode, 0)
        self.assertRaises(RuntimeError, proc.spawn, 'true')
        proc.close()
        proc.spawn('false')
        proc.wait()
        self.assertEqual(proc.returncode, 1)
        proc.close()

    def test_spawn_shell(self):
        # Ensure spawning a child with shell=True works.
        proc = Process()
        proc.spawn('true', shell=True)
        proc.wait()
        self.assertEqual(proc.returncode, 0)
        proc.close()
        proc.spawn('false', shell=True)
        proc.wait()
        self.assertEqual(proc.returncode, 1)
        proc.close()

    def test_spawn_args(self):
        # Ensure that passing arguments to our child works.
        proc = Process()
        proc.spawn(['exitn', '1', '2', '3'])
        proc.wait()
        self.assertEqual(proc.returncode, 6)
        proc.close()

    def test_spawn_shell_args(self):
        # Ensure that passing arguments to our child works with shell=True.
        proc = Process()
        proc.spawn('exitn 1 2 3', shell=True)
        proc.wait()
        self.assertEqual(proc.returncode, 6)
        proc.close()

    def test_spawn_executable(self):
        # Ensure that spawn honors the executable argument.
        proc = Process()
        proc.spawn(['exit', '1'], executable='true')
        proc.wait()
        self.assertEqual(proc.returncode, 0)
        proc.close()

    def test_spawn_shell_executable(self):
        # Ensure that spawn honors the executable argument with shell=True.
        proc = Process()
        proc.spawn(['exit 1'], shell=True, executable='echo', stdout=PIPE)
        output = proc.stdout.readline().split()
        proc.wait()
        self.assertEqual(proc.returncode, 0)
        self.assertEqual(output[0], b'-c' if hasattr(os, 'fork') else b'/c')
        self.assertEqual(output[1], b'exit')
        self.assertEqual(output[2], b'1')
        proc.close()

    def test_exit(self):
        # Ensure that the child's exist status is correctly reported.
        proc = Process()
        proc.spawn(['exitn', '5'])
        proc.wait()
        self.assertEqual(proc.returncode, 5)
        proc.close()

    def test_spawn_cwd(self):
        # Ensure that the "cwd" argument to spawn is effective.
        proc = Process()
        encoding = locale.getpreferredencoding()
        curdir = os.getcwd()
        tempdir = self.tempdir
        self.assertNotEqual(tempdir, curdir)
        proc.spawn('pwd', stdout=PIPE)
        childdir = proc.stdout.readline().rstrip().decode(encoding)
        proc.wait()
        self.assertEqual(proc.returncode, 0)
        self.assertEqual(childdir, curdir)
        proc.close()
        proc = Process()
        proc.spawn('pwd', stdout=PIPE, cwd=tempdir)
        childdir = proc.stdout.readline().rstrip().decode(encoding)
        proc.wait()
        self.assertEqual(proc.returncode, 0)
        self.assertEqual(childdir, tempdir)
        proc.close()

    def test_spawn_env(self):
        # Ensure that the "env" argument to spawn is effective.
        proc = Process()
        encoding = locale.getpreferredencoding()
        env = os.environ.copy()
        env['FOO'] = 'Bar'
        self.assertNotEqual(os.environ.get('FOO'), env['FOO'])
        proc.spawn(['echo', '$FOO'], stdout=PIPE, env=env)
        value = proc.stdout.readline().rstrip().decode(encoding)
        proc.wait()
        self.assertEqual(proc.returncode, 0)
        self.assertEqual(value, env['FOO'])
        proc.close()

    def test_returncode(self):
        # Ensure that the returncode attribute gets set when the child exits.
        proc = Process()
        proc.spawn(['sleep', '0.2'])
        tries = 0
        while True:
            if proc.returncode is not None:
                break
            tries += 1
            gruvi.sleep(0.02)
        self.assertEqual(proc.returncode, 0)
        self.assertGreater(tries, 5)
        proc.close()

    def test_wait(self):
        # Ensure that wait() waits for the child to exit.
        proc = Process()
        t0 = time.time()
        proc.spawn(['sleep', '0.2'])
        proc.wait()
        t1 = time.time()
        self.assertGreater(t1-t0, 0.2)
        self.assertEqual(proc.returncode, 0)
        proc.close()

    def test_wait_timeout(self):
        # Ensure that wait() honors its "timeout" argument.
        proc = Process()
        proc.spawn(['sleep', '10'])
        self.assertRaises(Timeout, proc.wait, 0.1)
        proc.terminate()
        proc.wait()
        proc.close()

    def test_stdout(self):
        # Ensure that it's possible to capure stdout using stdout=PIPE
        proc = Process()
        proc.spawn('catn', stdin=PIPE, stdout=PIPE)
        proc.stdin.write(b'Foo\n')
        proc.stdin.close()
        self.assertEqual(proc.stdout.readline(), b'Foo\n')
        self.assertEqual(proc.stdout.readline(), b'')
        proc.wait()
        self.assertEqual(proc.returncode, 0)
        proc.close()

    def test_stderr(self):
        # Ensure that it's possible to capure stderr using stderr=PIPE
        proc = Process()
        proc.spawn(['catn', '0', '2'], stdin=PIPE, stdout=PIPE, stderr=PIPE)
        proc.stdin.write(b'Foo\n')
        proc.stdin.close()
        self.assertEqual(proc.stderr.readline(), b'Foo\n')
        self.assertEqual(proc.stderr.readline(), b'')
        self.assertEqual(proc.stdout.readline(), b'')
        proc.wait()
        self.assertEqual(proc.returncode, 0)
        proc.close()

    def test_devnull(self):
        # Ensure that using stdout=DEVNULL doesn't produce any output.
        proc = Process()
        proc.spawn('catn', stdin=PIPE, stdout=DEVNULL)
        proc.stdin.write(b'Foo\n')
        proc.stdin.close()
        self.assertIsNone(proc.stdout)
        proc.wait()
        self.assertEqual(proc.returncode, 0)
        proc.close()

    def test_stdio_encoding(self):
        # Ensure that passing encoding=xxx to the constructor works.
        encoding = locale.getpreferredencoding()
        proc = Process(encoding=encoding)
        proc.spawn('catn', stdin=PIPE, stdout=PIPE)
        proc.stdin.write(u'20 \u20ac\n')
        proc.stdin.close()
        self.assertEqual(proc.stdout.readline(), u'20 \u20ac\n')
        self.assertEqual(proc.wait(), 0)
        proc.close()

    def test_spawn_unicode_args(self):
        # Ensure that it's possible to spawn a child with unicode arguments.
        proc = Process()
        env = os.environ.copy()
        # Ensure to have a capable sys.stdout encoding.
        env['PYTHONIOENCODING'] = 'utf-8'
        proc.spawn(['echo', 'foo', u'\u20ac'], stdout=PIPE, env=env)
        # Unsure why a \x00 is present at the end
        line = proc.stdout.readline().rstrip()
        self.assertEqual(line, u'foo \u20ac'.encode('utf-8'))
        proc.wait()
        proc.close()

    def test_spawn_unicode_env(self):
        # Ensure that it's possible to spawn a child with a unicode environment.
        proc = Process()
        env = os.environ.copy()
        env['PYTHONIOENCODING'] = 'utf-8'
        env['FOO'] = u'foo \u20ac'
        proc.spawn(['echo', '$FOO'], stdout=PIPE, env=env)
        line = proc.stdout.readline().rstrip()
        self.assertEqual(line, u'foo \u20ac'.encode('utf-8'))
        proc.wait()
        proc.close()

    def test_timeout(self):
        # Ensure that the timeout=xxx constructor argument works.
        proc = Process(timeout=0.1)
        proc.spawn('catn', stdin=PIPE, stdout=PIPE)
        self.assertRaises(Timeout, proc.stdout.readline)
        proc.stdin.write(b'foo\n')
        self.assertEqual(proc.stdout.readline(), b'foo\n')
        proc.stdin.close()
        self.assertEqual(proc.wait(), 0)
        proc.close()

    def test_inherit_handle(self):
        # Ensure that it's possible to pass a handle to the child.
        # Note: The "ipc" flag doubles as a read/write flag.
        hub = get_hub()
        handle = pyuv.Pipe(hub.loop, True)
        proc = Process()
        proc.spawn(python_script(['catn', '3']), extra_handles=[handle])
        stream = StreamClient()
        stream.connect(handle)
        stream.write(b'Foo\n')
        self.assertEqual(stream.readline(), b'Foo\n')
        stream.write_eof()
        self.assertEqual(stream.readline(), b'')
        stream.close()
        proc.wait()
        self.assertEqual(proc.returncode, 0)
        proc.close()

    def test_send_signal(self):
        # Ensure we can send a signal to our child using send_signal(0
        proc = Process()
        proc.spawn(['sleep', '1'])
        proc.send_signal(signal.SIGINT)
        proc.wait()
        self.assertEqual(proc.returncode, -signal.SIGINT)
        proc.close()

    def test_terminate(self):
        # Ensure that terminate() kills our child.
        proc = Process()
        proc.spawn(['sleep', '1'])
        proc.terminate()
        proc.wait()
        self.assertEqual(proc.returncode, -signal.SIGTERM)
        proc.terminate()  # should not error
        proc.close()

    def test_child_exited(self):
        # Ensure that the child_exited callback gets called.
        proc = Process()
        cbargs = []
        def on_child_exit(*args):
            cbargs.extend(args)
        proc.child_exited = on_child_exit
        proc.spawn(['sleep', '0'])
        proc.wait()
        self.assertEqual(proc.returncode, 0)
        self.assertEqual(cbargs[0], 0)
        self.assertEqual(cbargs[1], 0)
        proc.close()

    def test_send_data(self):
        # Test that sending a lot of data works.
        proc = Process()
        proc.spawn('catn', stdin=PIPE, stdout=PIPE)
        buf = b'x' * 1024
        nbytes = 10*1024*1024  # Send 10MB
        result = [0, 0]
        def writer():
            while result[0] < nbytes:
                towrite = min(1024, nbytes - result[0])
                proc.stdin.write(buf[:towrite])
                result[0] += towrite
            proc.stdin.flush()
            proc.stdin.write_eof()
        def reader():
            while True:
                read = len(proc.stdout.read1(1024))
                if read == 0:
                    break
                result[1] += read
        gruvi.spawn(writer)
        gruvi.spawn(reader).join()
        proc.wait()
        self.assertEqual(proc.returncode, 0)
        self.assertEqual(result[0], nbytes)
        self.assertEqual(result[1], nbytes)
        proc.close()

    def test_communicate(self):
        # Test that communicate() works
        proc = Process()
        proc.spawn('catn', stdin=PIPE, stdout=PIPE)
        buf = b'x' * 1024
        stdout, stderr = proc.communicate(buf)
        self.assertEqual(stdout, buf)
        self.assertEqual(len(stderr), 0)
        self.assertEqual(proc.returncode, 0)
        proc.close()

    def test_communicate_stderr(self):
        # Test that communicate() works with stderr
        proc = Process()
        proc.spawn(['catn', '0', '2'], stdin=PIPE, stderr=PIPE)
        buf = b'x' * 1024
        stdout, stderr = proc.communicate(buf)
        self.assertEqual(len(stdout), 0)
        self.assertEqual(stderr, buf)
        self.assertEqual(proc.returncode, 0)
        proc.close()

    def test_communicate_timeout(self):
        # Test that communicate() honors its "timeout" argument
        proc = Process()
        proc.spawn(['sleep', '10'], stdin=PIPE, stdout=PIPE)
        buf = b'x' * 1024
        self.assertRaises(Timeout, proc.communicate, buf, timeout=0.1)
        proc.terminate()
        proc.wait()
        proc.close()

    def test_no_child(self):
        # Test method behavior when there is no child.
        proc = Process()
        self.assertIsNone(proc.pid)
        self.assertRaises(RuntimeError, proc.send_signal, signal.SIGTERM)
        self.assertRaises(RuntimeError, proc.terminate)
        self.assertRaises(RuntimeError, proc.wait)
        self.assertRaises(RuntimeError, proc.communicate)
        proc.close()


if __name__ == '__main__':
    unittest.main()



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import sys
import threading
import functools


# Some compatibility stuff that is not in six.

PY3 = sys.version_info[0] == 3


# Provide a get_thread_ident() that is the same on Python 2.x and 3.x

try:
    get_thread_ident = threading._get_ident
except AttributeError:
    get_thread_ident = threading.get_ident


# Support for re-raising stored exceptions. In our callback based code it is
# very common that we store an exception in a callback as an instance
# attribute, and then raise that from the consumer API. This is problematic on
# Python3 because the traceback in Py3K is attached to the exception itself (as
# the __traceback__ attribute). The impact is that i) the traceback will also
# be stored and never freed, and ii) if the exception is re-raised multiple times
# a new traceback will be appended for every raise.
#
# The solution that we're using is is to have a saved_exc() call that returns a
# new copy of the exception on Python3.

if PY3:
    def saved_exc(exc):
        if isinstance(exc, type):
            return exc
        return type(exc)(*exc.args)
else:
    def saved_exc(exc):
        return exc


# Support write_through for TextIOWrapper on Python 2.x. The write_through
# argument first appeared in Python 3.3.

def _textiowrapper_write(self, buf):
    ret = self._orig_write(buf)
    self.flush()
    return ret

def _textiowrapper_writelines(self, seq):
    ret = self._orig_writelines(seq)
    self.flush()
    return ret

def make_textiowrapper_writethrough(wrapper):
    """Enable write_through for a TextIOWrapper."""
    wrapper._orig_write = wrapper.write
    wrapper.write = functools.partial(_textiowrapper_write, wrapper)
    wrapper._orig_writelines = wrapper.writelines
    wrapper.writelines = functools.partial(_textiowrapper_writelines, wrapper)



"""
Provides data marshalling to and from the DBus wire format

@author: Tom Cocagne
"""

from __future__ import absolute_import, print_function

import struct
import re
import six
import codecs

from .error import MarshallingError


invalid_obj_path_re = re.compile('[^a-zA-Z0-9_/]')
if_re               = re.compile('[^A-Za-z0-9_.]')
bus_re              = re.compile('[^A-Za-z0-9_.\-:]')
mbr_re              = re.compile('[^A-Za-z0-9_]')
dot_digit_re        = re.compile('\.\d')


#                Name      Type code   Alignment
dbus_types = [ ('BYTE',        'y',     1),
               ('BOOLEAN',     'b',     4),
               ('INT16',       'n',     2),
               ('UINT16',      'q',     2),
               ('INT32',       'i',     4),
               ('UINT32',      'u',     4),
               ('INT64',       'x',     8),
               ('UINT64',      't',     8),
               ('DOUBLE',      'd',     8),
               ('STRING',      's',     4), # (4-byte align for length)
               ('OBJECT_PATH', 'o',     4), # (4-byte align for length)
               ('SIGNATURE',   'g',     1),
               ('ARRAY',       'a',     4), # (4-byte align for length)
               ('STRUCT',      '(',     8),
               ('VARIANT',     'v',     1), # (1-byte align for signature)
               ('DICT_ENTRY',  '{',     8),
               ('UNIX_FD',     'h',     4)
               ]


class Byte(int):
    """
    Used during Variant serialization to ensure that this type is
    encoded rather than the generic Python type
    """
    dbusSignature = 'y'
    
class Boolean(int):
    """
    Used during Variant serialization to ensure that this type is
    encoded rather than the generic Python type
    """
    dbusSignature = 'b'
    
class Int16(int):
    """
    Used during Variant serialization to ensure that this type is
    encoded rather than the generic Python type
    """
    dbusSignature = 'n'
    
class UInt16(int):
    """
    Used during Variant serialization to ensure that this type is
    encoded rather than the generic Python type
    """
    dbusSignature = 'q'
    
class Int32(int):
    """
    Used during Variant serialization to ensure that this type is
    encoded rather than the generic Python type
    """
    dbusSignature = 'i'
    
class UInt32(int):
    """
    Used during Variant serialization to ensure that this type is
    encoded rather than the generic Python type
    """
    dbusSignature = 'u'
    
class Int64(int):
    """
    Used during Variant serialization to ensure that this type is
    encoded rather than the generic Python type
    """
    dbusSignature = 'x'
    
class UInt64(int):
    """
    Used during Variant serialization to ensure that this type is
    encoded rather than the generic Python type
    """
    dbusSignature = 't'
    
class Signature (str):
    """
    Used during Variant serialization to ensure that this type is
    encoded rather than the generic Python type
    """
    dbusSignature = 'g'
    
class ObjectPath (str):
    """
    Used during Variant serialization to ensure that this type is
    encoded rather than the generic Python type
    """
    dbusSignature = 'o'


variantClassMap = { 'y' : Byte,
                    'b' : Boolean,
                    'n' : Int16,
                    'q' : UInt16,
                    'i' : Int32,
                    'u' : UInt32,
                    'x' : Int64,
                    't' : UInt64,
                    'g' : Signature,
                    'o' : ObjectPath }

def validateObjectPath(p):
    """
    Ensures that the provided object path conforms to the DBus standard.
    Throws a L{error.MarshallingError} if non-conformant

    @type p: C{string}
    @param p: A DBus object path
    """
    if not p.startswith('/'):
        raise MarshallingError('Object paths must begin with a "/"')
    if len(p) > 1 and p[-1] == '/':
        raise MarshallingError('Object paths may not end with "/"')
    if '//' in p:
        raise MarshallingError('"//" is not allowed in object paths"')
    if invalid_obj_path_re.search(p):
        raise MarshallingError('Invalid characters contained in object path')


def validateInterfaceName( n ):
    """
    Verifies that the supplied name is a valid DBus Interface name. Throws
    an L{error.MarshallingError} if the format is invalid

    @type n: C{string}
    @param n: A DBus interface name
    """
    try:
        if not '.' in n:
            raise Exception('At least two components required')
        if '..' in n:
            raise Exception('".." not allowed in interface names')
        if len(n) > 255:
            raise Exception('Name exceeds maximum length of 255')
        if n[0] == '.':
            raise Exception('Names may not begin with a "."')
        if n[0].isdigit():
            raise Exception('Names may not begin with a digit')
        if if_re.search(n):
            raise Exception('Names contains a character outside the set [A-Za-z0-9_.]')
        if dot_digit_re.search(n):
            raise Exception('No components of an interface name may begin with a digit')
    except Exception as e:
        raise MarshallingError('Invalid interface name "%s": %s' % (n, str(e)))


def validateErrorName( n ):
    try:
        validateInterfaceName( n )
    except MarshallingError as e:
        raise MarshallingError( str(e).replace( 'interface', 'error', 1 ) )

    
def validateBusName( n ):
    """
    Verifies that the supplied name is a valid DBus Bus name. Throws
    an L{error.MarshallingError} if the format is invalid

    @type n: C{string}
    @param n: A DBus bus name
    """
    try:
        if not '.' in n:
            raise Exception('At least two components required')
        if '..' in n:
            raise Exception('".." not allowed in bus names')
        if len(n) > 255:
            raise Exception('Name exceeds maximum length of 255')
        if n[0] == '.':
            raise Exception('Names may not begin with a "."')
        if n[0].isdigit():
            raise Exception('Names may not begin with a digit')
        if bus_re.search(n):
            raise Exception('Names contains a character outside the set [A-Za-z0-9_.\-:]')
        if not n[0] == ':' and dot_digit_re.search(n):
            raise Exception('No coponents of an interface name may begin with a digit')
    except Exception as e:
        raise MarshallingError('Invalid bus name "%s": %s' % (n, str(e)))

    
def validateMemberName( n ):
    """
    Verifies that the supplied name is a valid DBus member name. Throws
    an L{error.MarshallingError} if the format is invalid

    @type n: C{string}
    @param n: A DBus member name
    """
    try:
        if len(n) < 1:
            raise Exception('Name must be at least one byte in length')
        if len(n) > 255:
            raise Exception('Name exceeds maximum length of 255')
        if n[0].isdigit():
            raise Exception('Names may not begin with a digit')
        if mbr_re.search(n):
            raise Exception('Names contains a character outside the set [A-Za-z0-9_]')
    except Exception as e:
        raise MarshallingError('Invalid member name "%s": %s' % (n, str(e)))

    

# XXX: This could be made *much* smarter (handle objects and recursive structures)
def sigFromPy( pobj ):
    """
    Returns the DBus signature type for the argument. If the argument is an
    instance of one of the type wrapper classes, the exact type signature
    corresponding to the wrapper class will be used. If the object has a
    variable named 'dbusSignature', the value of that variable will be
    used. Otherwise, a generic type will be used (i.e "i" for a Python int)

    @rtype: C{string}
    @returns: The DBus signature for the supplied Python object
    """
    sig = getattr(pobj, 'dbusSignature', None)
    
    if sig is not None:
        return sig
    
    elif isinstance(pobj,        int): return 'i'
    elif isinstance(pobj, six.integer_types): return 'x'
    elif isinstance(pobj,      float): return 'd'
    elif isinstance(pobj, six.string_types): return 's'
    
    elif isinstance(pobj,       list):
        vtype = type(pobj[0])
        same = True
        for v in pobj[1:]:
            if not vtype is type(v):
                same = False
        if same:
            return 'a' + sigFromPy(pobj[0])
        else:
            return 'av'
    
    elif isinstance(pobj,       dict):
        same = True
        vtype = None
        for k,v in six.iteritems(pobj):
            if vtype is None:
                vtype = type(v)
            elif not vtype is type(v):
                same = False
        if same:
            return 'a{' + sigFromPy(k) + sigFromPy(v) + '}'
        else:
            return 'a{' + sigFromPy(k) + 'v}'
    
    else:
        raise MarshallingError('Invalid Python type for variant: ' + repr(pobj))

    
#-------------------------------------------------------------------------------
#                          Marshalling Functions
# Padding:
#    - All data types must be padded to the correct alignment
#    - All padding bytes must be nul
#
padding = { 0 : b'\0' * 0,
            1 : b'\0' * 1,
            2 : b'\0' * 2,
            3 : b'\0' * 3,
            4 : b'\0' * 4,
            5 : b'\0' * 5,
            6 : b'\0' * 6,
            7 : b'\0' * 7 }

def genpad( align ):
    return lambda x : padding[ x % align and (align - x%align) or 0 ]

pad = dict()

for name, tcode, align in dbus_types:
    pad[ tcode ] = genpad(align)

pad['header'] = genpad(8)


#-------------------------------------------------------------------------------
#                    Signature Generator/Iterator
#
def genCompleteTypes( compoundSig ):
    """
    Generator function used to iterate over each complete,
    top-level type contained in in a signature. Ex::
      "iii"       => [ 'i', 'i',       'i' ]
      "i(ii)i"    => [ 'i', '(ii)',    'i' ]
      "i(i(ii))i" => [ 'i', '(i(ii))', 'i' ]
    """

    i     = 0
    start = 0
    end   = len(compoundSig)

    def find_end( idx, b, e ):
        depth = 1
        while idx < end:
            subc = compoundSig[idx]
            if subc == b:
                depth += 1
            elif subc == e:
                depth -= 1
                if depth == 0:
                    return idx
            idx += 1
    
    while i < end:
        c = compoundSig[i]
        
        if c == '(':
            x = find_end(i+1, '(', ')')
            yield compoundSig[i:x+1]
            i = x

        elif c == '{':
            x = find_end(i+1, '{', '}')
            yield compoundSig[i:x+1]
            i = x

        elif c == 'a':
            start = i
            g = genCompleteTypes( compoundSig[i+1:] )
            ct = six.next(g)
            i += len(ct)
            yield 'a' + ct
            
        else:
            yield c
        i += 1



#-------------------------------------------------------------------------------
#                          Marshalling Functions
# General:
#    - All values must be padded to proper alignment
#    - Pad bytes must be zero
#
# BOOLEAN:
#    - Only 1 & 0 are valid
#
# DICT_ENTRY:
#    - Identical to STRUCT
#
# Message:
#    - Max length of header, body and all padding is 2^27
#
# Message Header:
#    - Must be padded to a multiple of 8 bytes
#    - Fixed signature: "yyyyuua(yv)"
#        *  1 BYTE:   Endian flag. 'l' for little, 'B' for big
#        *  2 BYTE:   Message type enum
#        *  3 BYTE:   Bit Flags
#        *  4 BYTE:   Major protocol version (1 currently)
#        *  5 UINT32: Body Length (begins after header padding)
#        *  6 UINT32: Message serial number (must not be zero)
#        *  7 Array:  zero or more header fields. Msg type determines
#                     which entries are required
#
# Message Body:
#    - Begins on 8-byte boundary
#    - Not padded to a required byte alignment
#
#

def marshal_byte( ct, var, start_byte, lendian ):
    return 1, [ struct.pack( lendian and '<B' or '>B', var) ]

def marshal_boolean( ct, var, start_byte, lendian ):
    return 4, [ struct.pack( lendian and '<I' or '>I', 1 if var else 0) ]

def marshal_int16( ct, var, start_byte, lendian ):
    return 2, [ struct.pack( lendian and '<h' or '>h', var) ]

def marshal_uint16( ct, var, start_byte, lendian ):
    return 2, [ struct.pack( lendian and '<H' or '>H', var) ]

def marshal_int32( ct, var, start_byte, lendian ):
    return 4, [ struct.pack( lendian and '<i' or '>i', var) ]

def marshal_uint32( ct, var, start_byte, lendian ):
    return 4, [ struct.pack( lendian and '<I' or '>I', var) ]

def marshal_int64( ct, var, start_byte, lendian ):
    return 8, [ struct.pack( lendian and '<q' or '>q', var) ]

def marshal_uint64( ct, var, start_byte, lendian ):
    return 8, [ struct.pack( lendian and '<Q' or '>Q', var) ]

def marshal_double( ct, var, start_byte, lendian ):
    return 8, [ struct.pack( lendian and '<d' or '>d', var) ]


# STRING:
#    - *must* be valid UTF-8, nul terminated with no embedded nuls
#    format:
#       1 - UINT32 length in bytes (excluding terminating nul)
#       2 - string data (no embedded nuls)
#       3 - terminating nul byte
#
def marshal_string( ct, var, start_byte, lendian ):
    if not isinstance(var, six.string_types):
        raise MarshallingError('Required string. Received: ' + repr(var))
    if var.find('\0') != -1:
        raise MarshallingError('Embedded nul characters are not allowed within DBus strings')
    var = codecs.encode(var, 'utf-8')
    return 4 + len(var) + 1, [ struct.pack( lendian and '<I' or '>I', len(var)), var, b'\0' ]


# OBJECT_PATH:
#    - Identical to string
#    
def marshal_object_path( ct, var, start_byte, lendian ):
    validateObjectPath(var)
    return marshal_string( ct, var, start_byte, lendian )


# SIGNATURE:
#    - Ends with nul byte
#    - List of complete types. No partial types permitted
#    - Max signature length is 255
#    format:
#       1 - Single byte length
#       2 - Valid signature string
#       3 - terminating nul byte
def marshal_signature( ct, var, start_byte, lendian ):
    # XXX validate signature
    var = codecs.encode(var, 'ascii')
    return 2 + len(var), [struct.pack(lendian and '<B' or '>B', len(var)), var, b'\0']


# ARRAY:
#    - Max length is 2^26
#    format:
#       1 - UINT32 length of array data (does not include alignment padding)
#       2 - Padding to required alignment of contained data type
#       3 - each array element
def marshal_array( ct, var, start_byte, lendian ):
    chunks   = list()
    data_len = 0
    tsig     = ct[1:]   # strip of leading 'a'
    tcode    = tsig[0] # type of array element

    start_byte += 4 # for array size
    
    initial_padding = pad[tcode]( start_byte )

    if initial_padding:
        start_byte += len(initial_padding)
        chunks.append( initial_padding )

    if isinstance(var, (list, tuple)):
        arr_list = var
    elif isinstance(var, dict):
        arr_list = [ tpl for tpl in six.iteritems(var) ]
    else:
        raise MarshallingError('List, Tuple, or Dictionary required for DBus array. Received: ' + repr(var))

    for item in arr_list:

        padding = pad[tcode]( start_byte )

        if padding:
            start_byte += len(padding)
            data_len   += len(padding)
            chunks.append( padding )
        
        nbytes, vchunks = marshallers[ tcode ]( tsig, item, start_byte, lendian )

        start_byte += nbytes
        data_len   += nbytes
        
        chunks.extend( vchunks )
        

    chunks.insert(0, struct.pack( lendian and '<I' or '>I', data_len))

    return 4 + len(initial_padding) + data_len, chunks
    

# STRUCT:
#    - Must start on 8 byte boundary
#    - Content consists of each field marshaled in sequence
#
def marshal_struct( ct, var, start_byte, lendian ):
    return marshal( ct[1:-1], var, start_byte, lendian )


marshal_dictionary = marshal_struct


# VARIANT:
#    - Signature must contain only a single, complete type
#    format:
#       1 - Marshaled SIGNATURE
#       2 - Any required padding to align the type specified in the signature
#       3 - Marshaled value
def marshal_variant( ct, var, start_byte, lendian ):
    # XXX: ensure only a single, complete type is in the siguature
    bstart = start_byte
    
    vsig = sigFromPy(var)
    
    nbytes, chunks = marshal_signature( ct, sigFromPy(var), start_byte, lendian )

    start_byte += nbytes

    padding = pad[vsig[0]]( start_byte )
        
    if padding:
        start_byte += len(padding)
        chunks.append( padding )
        
    rnbytes, rchunks = marshal( vsig, [var], start_byte, lendian )

    start_byte += rnbytes
    chunks.extend( rchunks )
    
    return start_byte - bstart, chunks



marshallers = { 'y' : marshal_byte,
                'b' : marshal_boolean,
                'n' : marshal_int16,
                'q' : marshal_uint16,
                'i' : marshal_int32,
                'u' : marshal_uint32,
                'x' : marshal_int64,
                't' : marshal_uint64,
                'd' : marshal_double,
                's' : marshal_string,
                'o' : marshal_object_path,
                'g' : marshal_signature,
                'a' : marshal_array,
                '(' : marshal_struct,
                'v' : marshal_variant,
                '{' : marshal_dictionary,
                'h' : marshal_uint32 }


def marshal( compoundSignature, variableList, startByte = 0, lendian=True ):
    """
    Encodes the Python objects in variableList into the DBus wire-format
    matching the supplied compoundSignature. This function retuns a list of
    binary strings is rather than a single string to simplify the recursive
    marshalling algorithm. A single string may be easily obtained from the
    result via: ''.join(list_of_binary_strings)
    
    @type compoundSignature: C{string}
    @param compoundSignature: DBus signature specifying the types of the
                              variables to encode

    @type variableList: C{list}
    @param variableList: List of variables to encode (length of the list
                         must exactly match the number of variables specified
                         in compoundSignature


    @type startByte: C{int}
    @param startByte: Used during recursive marshalling to ensure data
                      alignment requirements are met

    @type lendian: C{bool}
    @param lendian: True if the data should be serialized in
                    little-endian format

    
    @returns: (number_of_encoded_bytes, list_of_binary_strings)
    """
    chunks = list()
    bstart = startByte

    if hasattr(variableList, 'dbusOrder'):
        order = getattr(variableList, 'dbusOrder')
        variableList = [ getattr(variableList, attr_name) for attr_name in order ]

    for ct, var in zip(genCompleteTypes( compoundSignature ), variableList):
        tcode   = ct[0]
        padding = pad[tcode]( startByte )
        
        if padding:
            startByte += len(padding)
            chunks.append( padding )
        
        nbytes, vchunks = marshallers[ tcode ]( ct, var, startByte, lendian )

        startByte += nbytes
        
        chunks.extend( vchunks )
        

    return startByte - bstart, chunks



#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
#                         Unmarshalling Functions
#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------


def unmarshal_byte(ct, data, offset, lendian):
    return 1, struct.unpack_from( lendian and '<B' or '>B', data, offset)[0]

def unmarshal_boolean(ct, data, offset, lendian):
    return 4, struct.unpack_from( lendian and '<I' or '>I', data, offset)[0] != 0

def unmarshal_int16(ct, data, offset, lendian):
    return 2, struct.unpack_from( lendian and '<h' or '>h', data, offset)[0]

def unmarshal_uint16(ct, data, offset, lendian):
    return 2, struct.unpack_from( lendian and '<H' or '>H', data, offset)[0]

def unmarshal_int32(ct, data, offset, lendian):
    return 4, struct.unpack_from( lendian and '<i' or '>i', data, offset)[0]

def unmarshal_uint32(ct, data, offset, lendian):
    return 4, struct.unpack_from( lendian and '<I' or '>I', data, offset)[0]

def unmarshal_int64(ct, data, offset, lendian):
    return 8, struct.unpack_from( lendian and '<q' or '>q', data, offset)[0]

def unmarshal_uint64(ct, data, offset, lendian):
    return 8, struct.unpack_from( lendian and '<Q' or '>Q', data, offset)[0]

def unmarshal_double(ct, data, offset, lendian):
    return 8, struct.unpack_from( lendian and '<d' or '>d', data, offset)[0]


# STRING:
#    - *must* be valid UTF-8, nul terminated with no embedded nuls
#    format:
#       1 - UINT32 length in bytes (excluding terminating nul)
#       2 - string data (no embedded nuls)
#       3 - terminating nul byte
#
def unmarshal_string(ct, data, offset, lendian):
    slen = struct.unpack_from( lendian and '<I' or '>I', data, offset)[0]
    s = codecs.decode(data[ offset + 4 :  offset + 4 + slen ], 'utf-8')
    return 4 + slen + 1, s
    

# OBJECT_PATH:
#    - Identical to string
#    
unmarshal_object_path = unmarshal_string


# SIGNATURE:
#    - Ends with nul byte
#    - List of complete types. No partial types permitted
#    - Max signature length is 255
#    format:
#       1 - Single byte length
#       2 - Valid signature string
#       3 - terminating nul byte
def unmarshal_signature(ct, data, offset, lendian):
    slen = struct.unpack_from( lendian and '<B' or '>B', data, offset)[0]
    s = codecs.decode(data[ offset + 1 : offset + 1 + slen ], 'ascii')
    return 1 + slen + 1, s
    

# ARRAY:
#    - Max length is 2^26
#    format:
#       1 - UINT32 length of array data (does not include alignment padding)
#       2 - Padding to required alignment of contained data type
#       3 - each array element
def unmarshal_array(ct, data, offset, lendian):
    start_offset = offset
    values       = list()
    data_len     = struct.unpack_from( lendian and '<I' or '>I', data, offset)[0]
    tsig         = ct[1:]  # strip of leading 'a'
    tcode        = tsig[0] # type of array element
    
    offset += 4                         # 4-byte data length
    offset += len(pad[tcode]( offset )) # padding length
    
    end_offset = offset + data_len
    
    while offset < end_offset:

        offset += len(pad[tcode](offset))
        
        nbytes, value = unmarshallers[ tcode ]( tsig, data, offset, lendian )

        offset += nbytes
        values.append( value )

    if not offset == end_offset:
        raise MarshallingError('Invalid array encoding')

    if tcode == '{':
        d = dict()
        for item in values:
            d[ item[0] ] = item[1]
        values = d

    return offset - start_offset, values
    

# STRUCT:
#    - Must start on 8 byte boundary
#    - Content consists of each field marshaled in sequence
#
def unmarshal_struct(ct, data, offset, lendian):
    return unmarshal( ct[1:-1], data, offset, lendian )


unmarshal_dictionary = unmarshal_struct


# VARIANT:
#    - Signature must contain only a single, complete type
#    format:
#       1 - Marshaled SIGNATURE
#       2 - Any required padding to align the type specified in the signature
#       3 - Marshaled value
def unmarshal_variant(ct, data, offset, lendian):
    # XXX: ensure only a single, complete type is in the siguature
    start_offset = offset
    nsig, vsig = unmarshal_signature( ct, data, offset, lendian )

    offset += nsig
    
    offset += len(pad[vsig[0]](offset))

    nvar, value = unmarshal( vsig, data, offset, lendian )

    offset += nvar
    
    return offset - start_offset, value[0]




unmarshallers = { 'y' : unmarshal_byte,
                  'b' : unmarshal_boolean,
                  'n' : unmarshal_int16,
                  'q' : unmarshal_uint16,
                  'i' : unmarshal_int32,
                  'u' : unmarshal_uint32,
                  'x' : unmarshal_int64,
                  't' : unmarshal_uint64,
                  'd' : unmarshal_double,
                  's' : unmarshal_string,
                  'o' : unmarshal_object_path,
                  'g' : unmarshal_signature,
                  'a' : unmarshal_array,
                  '(' : unmarshal_struct,
                  'v' : unmarshal_variant,
                  '{' : unmarshal_dictionary,
                  'h' : unmarshal_uint32 }


def unmarshal( compoundSignature, data, offset = 0, lendian = True ):
    """
    Unmarshals DBus encoded data.

    @type compoundSignature: C{string}
    @param compoundSignature: DBus signature specifying the encoded value types

    @type data: C{string}
    @param data: Binary data

    @type offset: C{int}
    @param offset: Offset within data at which data for compoundSignature
                   starts (used during recursion)

    @type lendian: C{bool}
    @param lendian: True if data is encoded in little-endian format
    
    @returns: (number_of_bytes_decoded, list_of_values)
    """
    values       = list()
    start_offset = offset

    for ct in genCompleteTypes( compoundSignature ):
        tcode   = ct[0]
        offset += len(pad[tcode]( offset ))
                
        nbytes, value = unmarshallers[ tcode ]( ct, data, offset, lendian )

        offset += nbytes
        values.append( value )


    return offset - start_offset, values






"""
DBus errors
@author: Tom Cocagne
"""

class DBusException (Exception):
    """
    Base class for all expected DBus exceptions
    """
    pass


class DBusAuthenticationFailed (DBusException):
    pass


class MarshallingError (DBusException):
    """
    Thrown when errors are encountered by the marshalling/unmarshalling
    code
    """
    pass


class TimeOut (DBusException):
    """
    Used to indicate a timeout for remote DBus method calls that
    request a timeout value
    """
    pass


class IntrospectionFailed (DBusException):
    """
    Thrown if remote object introspection fails
    """
    pass


class RemoteError (DBusException):
    """
    Thrown in response to errors encountered during a remote
    method invocation. 

    @ivar errName: DBus error name
    @type errName: C{string}
    """
    message = ''

    def __init__(self, errName):
        self.errName = errName

    def __str__(self):
        return '%s: %s' % (self.errName, self.message) if self.message else self.errName

        
class FailedToAcquireName(DBusException):
    """
    Indicates a failed attempt to acquire a bus name
    """
    def __init__(self, new_name, returnCode):
        head = 'Failed to acquire bus name "%s": ' % (new_name,)
        if returnCode == 2:
            tail = 'Queued for name acquisition'
        elif returnCode == 3:
            tail = 'Name in use'
        else:
            tail = 'Unknown reason'

        DBusException.__init__(self, head + tail)
        
        self.returnCode = returnCode
        



# Gruvi example program: a "fortune" web service

import locale
import gruvi
import contextlib

def fortune_app(environ, start_response):
    print('New connection from {0}'.format(environ['REMOTE_ADDR']))
    proc = gruvi.Process(encoding=locale.getpreferredencoding())
    proc.spawn('fortune', stdout=gruvi.PIPE)
    with contextlib.closing(proc):
        fortune = proc.stdout.read()
        proc.wait()
    start_response('200 OK', [('Content-Type', 'text/plain; charset=utf-8')])
    return [fortune.encode('utf-8')]

server = gruvi.HttpServer(fortune_app)
server.listen(('localhost', 0))
for addr in server.addresses:
    print('Listen on {0}'.format(gruvi.saddr(addr)))

server.run()



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2015 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function

import os.path
from cffi import FFI

parent, _ = os.path.split(os.path.abspath(__file__))
topdir, _ = os.path.split(parent)


ffi = FFI()

ffi.set_source('jsonrpc_ffi', """
    #include "src/json_splitter.c"
    """, include_dirs=[topdir])

ffi.cdef("""
    #define OK ...
    #define INCOMPLETE ...
    #define ERROR ...

    struct split_context {
        const char *buf;
        int buflen;
        int offset;
        int error;
        ...;
    };

    int json_split(struct split_context *ctx);
""")


if __name__ == '__main__':
    ffi.compile()



#
# This file is part of Gruvi. Gruvi is free software available under the
# terms of the MIT license. See the file "LICENSE" that was provided
# together with this source file for the licensing terms.
#
# Copyright (c) 2012-2014 the Gruvi authors. See the file "AUTHORS" for a
# complete list.

from __future__ import absolute_import, print_function, division

import time
import unittest

from gruvi.jsonrpc import JsonRpcClient, JsonRpcServer
from gruvi import jsonrpc_ffi

from support import PerformanceTest
from test_jsonrpc import set_buffer, echo_app


class PerfJsonRpc(PerformanceTest):

    def perf_split_throughput(self):
        chunk = b'{' + b'x' * 1000 + b'}'
        buf = chunk * 10
        ctx = jsonrpc_ffi.ffi.new('struct split_context *')
        nbytes = 0
        t0 = t1 = time.time()
        while t1 - t0 < 1:
            set_buffer(ctx, buf)
            while ctx.offset != len(buf):
                error = jsonrpc_ffi.lib.json_split(ctx)
                self.assertEqual(error, 0)
            nbytes += len(buf)
            t1 = time.time()
        speed = nbytes / (t1 - t0) / (1024 * 1024)
        self.add_result(speed)

    def perf_message_throughput(self):
        server = JsonRpcServer(echo_app)
        server.listen(('127.0.0.1', 0))
        addr = server.addresses[0]
        client = JsonRpcClient()
        client.connect(addr)
        nrequests = 0
        t0 = t1 = time.time()
        while t1 - t0 < 1:
            result = client.call_method('echo', 'foo')
            self.assertEqual(result, ['foo'])
            nrequests += 1
            t1 = time.time()
        throughput = nrequests / (t1 - t0)
        self.add_result(throughput)
        server.close()
        client.close()


if __name__ == '__main__':
    unittest.defaultTestLoader.testMethodPrefix = 'perf'
    unittest.main()
